"""Unit tests for :mod:`darker.highlighting`"""

# pylint: disable=too-many-arguments,redefined-outer-name,unused-argument
# pylint: disable=protected-access

import os
import sys
from pathlib import Path
from shlex import shlex
from typing import Dict, Generator
from unittest.mock import patch

import pytest
from _pytest.fixtures import SubRequest
from pygments.token import Token

from darker.command_line import parse_command_line
from darker.highlighting import colorize, lexers, should_use_color


@pytest.fixture(scope="module")
def module_tmp_path(tmp_path_factory: pytest.TempPathFactory) -> Path:
    """Fixture for creating a module-scope temporary directory

    :param tmp_path_factory: The temporary path factory fixture from Pytest
    :return: The created directory path

    """
    return tmp_path_factory.mktemp("test_highlighting")


def unset_our_env_vars():
    """Unset the environment variables used in this test module"""
    os.environ.pop("PY_COLORS", None)
    os.environ.pop("NO_COLOR", None)
    os.environ.pop("FORCE_COLOR", None)


@pytest.fixture(scope="module", autouse=True)
def clean_environ():
    """Fixture for clearing unwanted environment variables

    The ``NO_COLOR`` and ``PY_COLORS`` environment variables are tested in this module,
    so we need to ensure they aren't already set.

    In all `os.environ` patching, we use our own low-level custom code instead of
    `unittest.mock.patch.dict` for performance reasons.

    """
    old = os.environ
    os.environ = old.copy()  # type: ignore  # noqa: B003
    unset_our_env_vars()

    yield

    os.environ = old  # noqa: B003


@pytest.fixture(params=["", "color = false", "color = true"])
def pyproject_toml_color(
    request: SubRequest, module_tmp_path: Path
) -> Generator[None, None, None]:
    """Parametrized fixture for the ``color =`` option in ``pyproject.toml``

    Creates three versions of ``pyproject.toml`` in ``module_tmp_path`` for a test
    function:

    Without the ``color =`` option::

        [tool.darker]

    With color turned off::

        [tool.darker]
        color = false

    With color turned on::

        [tool.darker]
        color = true

    :param request: The Pytest ``request`` object
    :param module_tmp_path: A temporary directory created by Pytest
    :yield: The ``color =`` option line in ``pyproject.toml``, or an empty string

    """
    pyproject_toml_path = module_tmp_path / "pyproject.toml"
    with pyproject_toml_path.open("w") as pyproject_toml:
        print(f"[tool.darker]\n{request.param}\n", file=pyproject_toml)

    yield request.param

    pyproject_toml_path.unlink()


@pytest.fixture(params=["", "tty"])
def tty(request: SubRequest) -> Generator[bool, None, None]:
    """Parametrized fixture for patching `sys.stdout.isatty`

    Patches `sys.stdout.isatty` to return either `False` or `True`. The parameter values
    are strings and not booleans in order to improve readability of parametrized tests.
    Custom patching for performance.

    :param request: The Pytest ``request`` object
    :yield: The patched `False` or `True` return value for `sys.stdout.isatty`

    """
    old_isatty = sys.stdout.isatty
    is_a_tty: bool = request.param == "tty"
    sys.stdout.isatty = lambda: is_a_tty  # type: ignore[assignment]

    yield is_a_tty

    sys.stdout.isatty = old_isatty  # type: ignore[assignment]


def _parse_environment_variables(definitions: str) -> Dict[str, str]:
    """Parse a ``"<var1>=<val1> <var2>=<val2>"`` formatted string into a dictionary

    :param definitions: The string to parse
    :return: The parsed dictionary

    """
    return dict(item.split("=") for item in shlex(definitions, punctuation_chars=" "))


@pytest.fixture(params=["", "NO_COLOR=", "NO_COLOR=foo"])
def env_no_color(request: SubRequest) -> Generator[Dict[str, str], None, None]:
    """Parametrized fixture for patching ``NO_COLOR``

    This fixture must come before `config_from_env_and_argv` in test function
    signatures.

    Patches the environment with or without the ``NO_COLOR`` environment variable. The
    environment is expressed as a space-separated string to improve readability of
    parametrized tests.

    :param request: The Pytest ``request`` object
    :yield: The patched items in the environment

    """
    os.environ.update(_parse_environment_variables(request.param))
    yield request.param
    unset_our_env_vars()


@pytest.fixture(params=["", "FORCE_COLOR=", "FORCE_COLOR=foo"])
def env_force_color(request: SubRequest) -> Generator[Dict[str, str], None, None]:
    """Parametrized fixture for patching ``FORCE_COLOR``

    This fixture must come before `config_from_env_and_argv` in test function
    signatures.

    Patches the environment with or without the ``FORCE_COLOR`` environment variable.
    The environment is expressed as a space-separated string to improve readability of
    parametrized tests.

    :param request: The Pytest ``request`` object
    :yield: The patched items in the environment

    """
    os.environ.update(_parse_environment_variables(request.param))
    yield request.param
    unset_our_env_vars()


@pytest.fixture(params=["", "PY_COLORS=0", "PY_COLORS=1"])
def env_py_colors(request: SubRequest) -> Generator[Dict[str, str], None, None]:
    """Parametrized fixture for patching ``PY_COLORS``

    This fixture must come before `config_from_env_and_argv` in test function
    signatures.

    Patches the environment with or without the ``PY_COLORS`` environment variable. The
    environment is expressed as a space-separated string to improve readability of
    parametrized tests.

    :param request: The Pytest ``request`` object
    :yield: The patched items in the environment

    """
    os.environ.update(_parse_environment_variables(request.param))
    yield request.param
    unset_our_env_vars()


@pytest.fixture
def uninstall_pygments() -> Generator[None, None, None]:
    """Fixture for uninstalling ``pygments`` temporarily"""
    mods = sys.modules.copy()
    del mods["darker.highlighting"]
    # cause an ImportError for `import pygments`:
    mods["pygments"] = None  # type: ignore[assignment]
    with patch.dict(sys.modules, mods, clear=True):

        yield


config_cache = {}


@pytest.fixture(params=[[], ["--no-color"], ["--color"]])
def config_from_env_and_argv(
    request: SubRequest, module_tmp_path: Path
) -> Generator[bool, None, None]:
    """Parametrized fixture for the ``--color`` / ``--no-color`` arguments

    Yields ``color`` configuration boolean values resulting from the current environment
    variables and a command line
    - with no color argument,
    - with the ``--color`` argument, and
    - with the ``--no--color`` argument.

    The ``NO_COLOR`` and ``PY_COLORS`` environment variables affect the resulting
    configuration, and must precede `config_from_env_and_argv` in test function
    signatures (if they are being used).

    :param request: The Pytest ``request`` object
    :param module_tmp_path: A temporary directory created by Pytest
    :yield: The list of arguments for the Darker command line

    """
    argv = request.param + [str(module_tmp_path / "dummy.py")]
    cache_key = (
        tuple(request.param),
        os.getenv("NO_COLOR"),
        os.getenv("FORCE_COLOR"),
        os.getenv("PY_COLORS"),
        (module_tmp_path / "pyproject.toml").read_bytes(),
    )
    if cache_key not in config_cache:
        _, config, _ = parse_command_line(argv)
        config_cache[cache_key] = config["color"]
    yield config_cache[cache_key]


def test_should_use_color_no_pygments(
    uninstall_pygments: None,
    pyproject_toml_color: str,
    env_no_color: str,
    env_force_color: str,
    env_py_colors: str,
    config_from_env_and_argv: bool,
    tty: bool,
) -> None:
    """Color output is never used if `pygments` is not installed

    All combinations of ``pyproject.toml`` options, environment variables and command
    line options affecting syntax highlighting are tested without `pygments`.

    """
    result = should_use_color(config_from_env_and_argv)

    assert result is False


@pytest.mark.parametrize(
    "config_from_env_and_argv, expect",
    [(["--no-color"], False), (["--color"], True)],
    indirect=["config_from_env_and_argv"],
)
def test_should_use_color_pygments_and_command_line_argument(
    pyproject_toml_color: str,
    env_no_color: str,
    env_force_color: str,
    env_py_colors: str,
    config_from_env_and_argv: bool,
    expect: bool,
    tty: bool,
) -> None:
    """--color / --no-color determines highlighting if `pygments` is installed

    All combinations of ``pyproject.toml`` options, environment variables and command
    line options affecting syntax highlighting are tested with `pygments` installed.

    """
    result = should_use_color(config_from_env_and_argv)

    assert result == expect


@pytest.mark.parametrize(
    "env_py_colors, expect",
    [("PY_COLORS=0", False), ("PY_COLORS=1", True)],
    indirect=["env_py_colors"],
)
@pytest.mark.parametrize("config_from_env_and_argv", [[]], indirect=True)
def test_should_use_color_pygments_and_py_colors(
    pyproject_toml_color: str,
    env_no_color: str,
    env_force_color: str,
    env_py_colors: str,
    config_from_env_and_argv: bool,
    tty: bool,
    expect: bool,
) -> None:
    """PY_COLORS determines highlighting when `pygments` installed and no cmdline args

    These tests are set up so that it appears as if
    - ``pygments`` is installed
    - there is no ``--color`` or `--no-color`` command line option

    All combinations of ``pyproject.toml`` options and environment variables affecting
    syntax highlighting are tested.

    """
    result = should_use_color(config_from_env_and_argv)

    assert result == expect


@pytest.mark.parametrize(
    "env_no_color, env_force_color, expect",
    [
        ("            ", "FORCE_COLOR=   ", "should_use_color() == True"),
        ("            ", "FORCE_COLOR=foo", "should_use_color() == True"),
        ("NO_COLOR=   ", "FORCE_COLOR=   ", "                          "),
        ("NO_COLOR=   ", "FORCE_COLOR=foo", "                          "),
        ("NO_COLOR=foo", "FORCE_COLOR=   ", "                          "),
        ("NO_COLOR=foo", "FORCE_COLOR=foo", "                          "),
    ],
    indirect=["env_no_color", "env_force_color"],
)
@pytest.mark.parametrize("config_from_env_and_argv", [[]], indirect=True)
def test_should_use_color_no_color_force_color(
    pyproject_toml_color: str,
    env_no_color: str,
    env_force_color: str,
    config_from_env_and_argv: bool,
    tty: bool,
    expect: str,
) -> None:
    """NO_COLOR/FORCE_COLOR determine highlighting in absence of PY_COLORS/cmdline args

    These tests are set up so that it appears as if
    - ``pygments`` is installed
    - the ``PY_COLORS`` environment variable is unset
    - there is no ``--color`` or `--no-color`` command line option

    All combinations of ``pyproject.toml`` options, ``NO_COLOR``, ``FORCE_COLOR`` and
    `sys.stdout.isatty` are tested.

    """
    result = should_use_color(config_from_env_and_argv)

    assert result == (expect == "should_use_color() == True")


@pytest.mark.parametrize("config_from_env_and_argv", [[]], indirect=True)
@pytest.mark.parametrize(
    "pyproject_toml_color, tty, expect",
    [
        # for readability, padded strings are used for parameters and the expectation
        ("             ", "   ", "                          "),
        ("             ", "tty", "should_use_color() == True"),
        ("color = false", "   ", "                          "),
        ("color = false", "tty", "                          "),
        ("color = true ", "   ", "should_use_color() == True"),
        ("color = true ", "tty", "should_use_color() == True"),
    ],
    indirect=["pyproject_toml_color", "tty"],
)
def test_should_use_color_pygments(
    pyproject_toml_color: str,
    tty: bool,
    config_from_env_and_argv: bool,
    expect: str,
) -> None:
    """Color output is enabled only if correct configuration options are in place

    These tests are set up so that it appears as if
    - ``pygments`` is installed (required for running the tests)
    - there is no ``--color`` or `--no-color`` command line option
    - the ``PY_COLORS`` environment variable isn't set to ``0`` or ``1`` (cleared by
      the auto-use ``clear_environ`` fixture)

    This test exercises the remaining combinations of ``pyproject.toml`` options and
    environment variables affecting syntax highlighting.

    """
    result = should_use_color(config_from_env_and_argv)

    assert result == (expect == "should_use_color() == True")


def test_colorize_with_no_color():
    """``colorize()`` does nothing when Pygments isn't available"""
    result = colorize("print(42)", "python", use_color=False)

    assert result == "print(42)"


@pytest.mark.parametrize(
    "text, lexer, use_color, expect",
    [
        (
            "except RuntimeError:",
            "python",
            True,
            {"\x1b[34mexcept\x1b[39;49;00m \x1b[36mRuntimeError\x1b[39;49;00m:"},
        ),
        ("except RuntimeError:", "python", False, {"except RuntimeError:"}),
        ("a = 1", "python", True, {"a = \x1b[34m1\x1b[39;49;00m"}),
        ("a = 1\n", "python", True, {"a = \x1b[34m1\x1b[39;49;00m\n"}),
        (
            "- a\n+ b\n",
            "diff",
            True,
            {
                # Pygments 2.4.0:
                "\x1b[31;01m- a\x1b[39;49;00m\n\x1b[32m+ b\x1b[39;49;00m\n",
                # Pygments 2.10.0:
                "\x1b[91m- a\x1b[39;49;00m\n\x1b[32m+ b\x1b[39;49;00m\n",
                # Pygments 2.11.2:
                "\x1b[91m- a\x1b[39;49;00m\x1b[37m\x1b[39;49;00m\n"
                + "\x1b[32m+ b\x1b[39;49;00m\x1b[37m\x1b[39;49;00m\n",
            },
        ),
        (
            "- a\n+ b\n",
            "diff",
            False,
            {"- a\n+ b\n"},
        ),
    ],
)
def test_colorize(text, lexer, use_color, expect):
    """``colorize()`` produces correct highlighted terminal output"""
    result = colorize(text, lexer, use_color)

    assert result in expect


@pytest.mark.parametrize(
    "text, expect",
    [
        (
            "path/to/file.py:42:",
            [
                (0, Token.Literal.String, "path/to/file.py"),
                (15, Token.Text, ":"),
                (16, Token.Literal.Number, "42"),
                (18, Token.Text, ":"),
                (19, Token.Literal.Number, ""),
            ],
        ),
        (
            "path/to/file.py:42:43:",
            [
                (0, Token.Literal.String, "path/to/file.py"),
                (15, Token.Text, ":"),
                (16, Token.Literal.Number, "42"),
                (18, Token.Text, ":"),
                (19, Token.Literal.Number, "43"),
                (21, Token.Text, ":"),
                (22, Token.Literal.Number, ""),
            ],
        ),
    ],
)
def test_location_lexer(text, expect):
    """Linter "path:linenum:colnum:" prefixes are lexed correctly"""
    location_lexer = lexers.LocationLexer()

    result = list(location_lexer.get_tokens_unprocessed(text))

    assert result == expect


@pytest.mark.parametrize(
    "text, expect",
    [
        (
            "  no coverage:     a = 1",
            [
                (0, Token.Literal.String, "  no coverage: "),
                (15, Token.Text, "    "),
                (19, Token.Name, "a"),
                (20, Token.Text, " "),
                (21, Token.Operator, "="),
                (22, Token.Text, " "),
                (23, Token.Literal.Number.Integer, "1"),
            ],
        ),
        (
            "C000 python(code) = not(highlighted)",
            [
                (0, Token.Error, "C000"),
                (4, Token.Literal.String, " "),
                (5, Token.Literal.String, "python(code)"),
                (17, Token.Literal.String, " "),
                (18, Token.Literal.String, "="),
                (19, Token.Literal.String, " "),
                (20, Token.Literal.String, "not(highlighted)"),
            ],
        ),
        (
            "C0000 Unused argument not highlighted",
            [
                (0, Token.Error, "C0000"),
                (5, Token.Literal.String, " "),
                (6, Token.Literal.String, "Unused argument "),
                (22, Token.Literal.String, "not"),
                (25, Token.Literal.String, " "),
                (26, Token.Literal.String, "highlighted"),
            ],
        ),
        (
            "E000 Unused variable not highlighted",
            [
                (0, Token.Error, "E000"),
                (4, Token.Literal.String, " "),
                (5, Token.Literal.String, "Unused variable "),
                (21, Token.Literal.String, "not"),
                (24, Token.Literal.String, " "),
                (25, Token.Literal.String, "highlighted"),
            ],
        ),
        (
            "E0000 Returning python_expression - is highlighted",
            [
                (0, Token.Error, "E0000"),
                (5, Token.Literal.String, " "),
                (6, Token.Literal.String, "Returning "),
                (16, Token.Name, "python_expression"),
                (33, Token.Literal.String, " "),
                (34, Token.Literal.String, "-"),
                (35, Token.Literal.String, " "),
                (36, Token.Literal.String, "is"),
                (38, Token.Literal.String, " "),
                (39, Token.Literal.String, "highlighted"),
            ],
        ),
        (
            "F000 Unused python_expression_highlighted",
            [
                (0, Token.Error, "F000"),
                (4, Token.Literal.String, " "),
                (5, Token.Literal.String, "Unused "),
                (12, Token.Name, "python_expression_highlighted"),
            ],
        ),
        (
            "F0000 Base type PythonClassHighlighted whatever",
            [
                (0, Token.Error, "F0000"),
                (5, Token.Literal.String, " "),
                (6, Token.Literal.String, "Base type "),
                (16, Token.Name, "PythonClassHighlighted"),
                (38, Token.Literal.String, " "),
                (39, Token.Literal.String, "whatever"),
            ],
        ),
        (
            "N000 imported from python.module.highlighted",
            [
                (0, Token.Error, "N000"),
                (4, Token.Literal.String, " "),
                (5, Token.Literal.String, "imported from "),
                (19, Token.Name, "python"),
                (25, Token.Operator, "."),
                (26, Token.Name, "module"),
                (32, Token.Operator, "."),
                (33, Token.Name, "highlighted"),
            ],
        ),
        (
            "N0000 (message-identifier) not-highlighted-in-the-middle",
            [
                (0, Token.Error, "N0000"),
                (5, Token.Literal.String, " "),
                (6, Token.Literal.String, "(message-identifier)"),
                (26, Token.Literal.String, " "),
                (27, Token.Literal.String, "not-highlighted-in-the-middle"),
            ],
        ),
        (
            "W000 at-the-end-highlight (message-identifier)",
            [
                (0, Token.Error, "W000"),
                (4, Token.Literal.String, " "),
                (5, Token.Literal.String, "at-the-end-highlight"),
                (25, Token.Literal.String, " "),
                (26, Token.Literal.String, "("),
                (27, Token.Error, "message-identifier"),
                (45, Token.Literal.String, ")"),
            ],
        ),
        (
            "W0000 four-digit-warning",
            [
                (0, Token.Error, "W0000"),
                (5, Token.Literal.String, " "),
                (6, Token.Literal.String, "four-digit-warning"),
            ],
        ),
        (
            "E00 two-digit-message-id-not-highlighted",
            [
                (0, Token.Text, ""),
                (0, Token.Literal.String, "E00"),
                (3, Token.Literal.String, " "),
                (4, Token.Literal.String, "two-digit-message-id-not-highlighted"),
            ],
        ),
        (
            "E00000 five-digit-message-id-not-highlighted",
            [
                (0, Token.Text, ""),
                (0, Token.Literal.String, "E00000"),
                (6, Token.Literal.String, " "),
                (7, Token.Literal.String, "five-digit-message-id-not-highlighted"),
            ],
        ),
    ],
)
def test_description_lexer(text, expect):
    """The description parts of linter output are lexed correctly"""
    description_lexer = lexers.DescriptionLexer()

    result = list(description_lexer.get_tokens_unprocessed(text))

    assert result == expect



"""Exceptions which are used in multiple Darker modules"""


class DependencyError(Exception):
    """Parent class for exceptions about problems with dependencies of Darker"""


class IncompatiblePackageError(DependencyError):
    """Raised if an incompatible version of a required or optional package is found"""


class MissingPackageError(DependencyError):
    """Raised if a required or optional package is needed but is missing"""



"""Miscellaneous utility functions"""

import io
import logging
import tokenize
from datetime import datetime
from itertools import chain
from pathlib import Path
from typing import Collection, Iterable, List, Tuple, Union

logger = logging.getLogger(__name__)

TextLines = Tuple[str, ...]


GIT_DATEFORMAT = "%Y-%m-%d %H:%M:%S.%f +0000"


def detect_newline(string: str) -> str:
    """Detect LF or CRLF newlines in a string by looking at the end of the first line"""
    first_lf_pos = string.find("\n")
    if first_lf_pos > 0 and string[first_lf_pos - 1] == "\r":
        return "\r\n"
    return "\n"


class TextDocument:
    """Store & handle a multi-line text document, either as a string or list of lines"""

    DEFAULT_ENCODING = "utf-8"
    DEFAULT_NEWLINE = "\n"

    def __init__(
        self,
        string: str = None,
        lines: Iterable[str] = None,
        encoding: str = DEFAULT_ENCODING,
        newline: str = DEFAULT_NEWLINE,
        mtime: str = "",
    ):
        self._string = string
        self._lines = None if lines is None else tuple(lines)
        self._encoding = encoding
        self._newline = newline
        self._mtime = mtime

    def string_with_newline(self, newline: str) -> str:
        """Return the document as a string, using the given newline sequence"""
        if self._string is None or detect_newline(self._string) != newline:
            return joinlines(self.lines or (), newline)
        return self._string

    @property
    def string(self) -> str:
        """Return the document as a string, converting and caching if necessary"""
        if self._string is None:
            self._string = self.string_with_newline(self.newline)
        return self._string

    @property
    def encoded_string(self) -> bytes:
        """Return the document as a bytestring, converting and caching if necessary"""
        return self.string.encode(self.encoding)

    @property
    def lines(self) -> TextLines:
        """Return the document as a list of lines converting and caching if necessary"""
        if self._lines is None:
            self._lines = tuple((self._string or "").splitlines())
        return self._lines

    @property
    def encoding(self) -> str:
        """Return the encoding used in the document"""
        return self._encoding

    @property
    def newline(self) -> str:
        """Return the newline character sequence used in the document"""
        return self._newline

    @property
    def mtime(self) -> str:
        """Return the last modification time of the document"""
        return self._mtime

    @classmethod
    def from_str(
        cls,
        string: str,
        encoding: str = DEFAULT_ENCODING,
        override_newline: str = None,
        mtime: str = "",
    ) -> "TextDocument":
        """Create a document object from a string

        :param string: The contents of the new text document
        :param encoding: The character encoding to be used when writing out the bytes
        :param override_newline: Replace existing newlines with the given newline string
        :param mtime: The modification time of the original file

        """
        newline = detect_newline(string)
        if override_newline and override_newline != newline:
            string = string.replace(newline, override_newline)
            newline = override_newline
        return cls(string, None, encoding=encoding, newline=newline, mtime=mtime)

    @classmethod
    def from_bytes(cls, data: bytes, mtime: str = "") -> "TextDocument":
        """Create a document object from a binary string

        :param data: The binary content of the new text document
        :param mtime: The modification time of the original file

        """
        srcbuf = io.BytesIO(data)
        encoding, lines = tokenize.detect_encoding(srcbuf.readline)
        if not lines:
            return cls(lines=[], encoding=encoding, mtime=mtime)
        return cls.from_str(data.decode(encoding), encoding=encoding, mtime=mtime)

    @classmethod
    def from_file(cls, path: Path) -> "TextDocument":
        """Create a document object by reading a text file

        Also store the last modification time of the file.

        """
        mtime = datetime.utcfromtimestamp(path.stat().st_mtime).strftime(GIT_DATEFORMAT)
        with path.open("rb") as srcbuf:
            return cls.from_bytes(srcbuf.read(), mtime)

    @classmethod
    def from_lines(
        cls,
        lines: Iterable[str],
        encoding: str = DEFAULT_ENCODING,
        newline: str = DEFAULT_NEWLINE,
        mtime: str = "",
    ) -> "TextDocument":
        """Create a document object from a list of lines

        The lines should be strings without trailing newlines. They should be encoded in
        UTF-8 unless a different encoding is specified with the ``encoding`` argument.

        """
        return cls(None, lines, encoding=encoding, newline=newline, mtime=mtime)

    def __eq__(self, other: object) -> bool:
        """Compare the equality two text documents, ignoring the modification times"""
        if not isinstance(other, TextDocument):
            return NotImplemented
        if not self._string and not self._lines:
            return not other._string and not other._lines
        return self.lines == other.lines

    def __repr__(self) -> str:
        """Return a Python representation of the document object"""
        encoding = (
            ""
            if self._encoding == self.DEFAULT_ENCODING
            else f", encoding={self.encoding!r}"
        )
        newline = (
            ""
            if self.newline == self.DEFAULT_NEWLINE
            else f", newline={self.newline!r}"
        )
        mtime = "" if not self._mtime else f", mtime={self._mtime!r}"
        return (
            f"{type(self).__name__}("
            f"[{len(self.lines)} lines]"
            f"{encoding}{newline}{mtime})"
        )


DiffChunk = Tuple[int, TextLines, TextLines]


def debug_dump(black_chunks: List[DiffChunk], edited_linenums: List[int]) -> None:
    """Print debug output. This is used in case of an unexpected failure."""
    if logger.getEffectiveLevel() > logging.DEBUG:
        return
    for offset, old_lines, new_lines in black_chunks:
        print(80 * "-")
        for delta, old_line in enumerate(old_lines):
            linenum = offset + delta
            edited = "*" if linenum in edited_linenums else " "
            print(f"{edited}-{linenum:4} {old_line}")
        for _, new_line in enumerate(new_lines):
            print(f" +     {new_line}")
    print(80 * "-")


def joinlines(lines: Iterable[str], newline: str = "\n") -> str:
    """Join a list of lines back, adding a linefeed after each line

    This is the reverse of ``str.splitlines()``.

    """
    return "".join(f"{line}{newline}" for line in lines)


def get_path_ancestry(path: Path) -> Iterable[Path]:
    reverse_parents = reversed(path.parents)
    if path.is_dir():
        return chain(reverse_parents, [path])
    else:
        return reverse_parents


def get_common_root(paths: Iterable[Path]) -> Path:
    """Find the deepest common parent directory of given paths"""
    resolved_paths = [path.resolve() for path in paths]
    parents = reversed(list(zip(*(get_path_ancestry(path) for path in resolved_paths))))
    for first_path, *other_paths in parents:
        if all(path == first_path for path in other_paths):
            return first_path
    raise ValueError(f"Paths have no common parent Git root: {resolved_paths}")


class Buf:
    def __init__(self, initial_bytes: bytes):
        self._buf = io.BytesIO(initial_bytes)
        self._line_starts: List[int] = []

    def __next__(self) -> str:
        self._line_starts.append(self._buf.tell())
        return next(self._buf).rstrip(b"\n").decode("utf-8")

    def __iter__(self) -> "Buf":
        return self

    def seek_line(self, lines_delta: int) -> None:
        if lines_delta > 0:
            raise NotImplementedError("Seeking forwards is not supported")
        for _ in range(-lines_delta):
            self._buf.seek(self._line_starts.pop())

    def next_line_startswith(self, prefix: Union[str, TextLines]) -> bool:
        """Peek at the next line, return ``True`` if it starts with the given prefix"""
        try:
            return next(self).startswith(prefix)
        except StopIteration:
            return False
        finally:
            self.seek_line(-1)


def glob_any(path: Path, patterns: Collection[str]) -> bool:
    """Return `True` if path matches any of the patterns

    Return `False` if there are no patterns to match.

    :param path: The file path to match
    :param patterns: The patterns to match against
    :return: `True` if at least one pattern matches

    """
    return any(path.glob(pattern) for pattern in patterns)



"""Unit tests for :mod:`darker.git`"""

# pylint: disable=redefined-outer-name,protected-access,too-many-arguments

import os
import re
from datetime import datetime, timedelta
from pathlib import Path
from subprocess import DEVNULL, PIPE, CalledProcessError, check_call  # nosec
from textwrap import dedent  # nosec
from typing import List, Union
from unittest.mock import call, patch

import pytest

from darker import git
from darker.tests.conftest import GitRepoFixture
from darker.tests.helpers import raises_or_matches
from darker.utils import GIT_DATEFORMAT, TextDocument


def test_tmp_path_sanity(tmp_path):
    """Make sure Pytest temporary directories aren't inside a Git repository"""
    try:
        result = git._git_check_output_lines(
            ["rev-parse", "--absolute-git-dir"], tmp_path, exit_on_error=False
        )
    except CalledProcessError as exc_info:
        if exc_info.returncode != 128 or not exc_info.stderr.startswith(
            "fatal: not a git repository"
        ):
            raise
    else:
        output = "\n".join(result)
        raise AssertionError(
            f"Temporary directory {tmp_path} for tests is not clean."
            f" There is a Git directory in {output}"
        )


@pytest.mark.parametrize(
    "revision_range, expect",
    [
        ("", None),
        ("..", ("", "..", "")),
        ("...", ("", "...", "")),
        ("a..", ("a", "..", "")),
        ("a...", ("a", "...", "")),
        ("a..b", ("a", "..", "b")),
        ("a...b", ("a", "...", "b")),
        ("..b", ("", "..", "b")),
        ("...b", ("", "...", "b")),
    ],
)
def test_commit_range_re(revision_range, expect):
    """Test for ``COMMIT_RANGE_RE``"""
    match = git.COMMIT_RANGE_RE.match(revision_range)
    if expect is None:
        assert match is None
    else:
        assert match is not None
        assert match.groups() == expect


def test_worktree_symbol():
    """Test for the ``WORKTREE`` symbol"""
    assert git.WORKTREE == ":WORKTREE:"


def test_git_get_mtime_at_commit():
    """darker.git.git_get_mtime_at_commit()"""
    with patch.object(git, "_git_check_output_lines"):
        git._git_check_output_lines.return_value = ["1609104839"]  # type: ignore

        result = git.git_get_mtime_at_commit(
            Path("dummy path"), "dummy revision", Path("dummy cwd")
        )
        assert result == "2020-12-27 21:33:59.000000 +0000"


@pytest.mark.kwparametrize(
    dict(
        revision=":WORKTREE:",
        expect_lines=("new content",),
        expect_mtime=lambda: datetime(2001, 9, 9, 1, 46, 40),
    ),
    dict(
        revision="HEAD",
        expect_lines=("modified content",),
        expect_mtime=datetime.utcnow,
    ),
    dict(
        revision="HEAD^",
        expect_lines=("original content",),
        expect_mtime=datetime.utcnow,
    ),
    dict(revision="HEAD~2", expect_lines=(), expect_mtime=False),
)
def test_git_get_content_at_revision(git_repo, revision, expect_lines, expect_mtime):
    """darker.git.git_get_content_at_revision()"""
    git_repo.add({"my.txt": "original content"}, commit="Initial commit")
    paths = git_repo.add({"my.txt": "modified content"}, commit="Initial commit")
    paths["my.txt"].write_bytes(b"new content")
    os.utime(paths["my.txt"], (1000000000, 1000000000))

    result = git.git_get_content_at_revision(
        Path("my.txt"), revision, cwd=Path(git_repo.root)
    )

    assert result.lines == expect_lines
    if expect_mtime:
        mtime_then = datetime.strptime(result.mtime, GIT_DATEFORMAT)
        difference = expect_mtime() - mtime_then
        assert timedelta(0) <= difference < timedelta(seconds=6)
    else:
        assert result.mtime == ""
    assert result.encoding == "utf-8"


@pytest.mark.parametrize(
    "revision_range, expect",
    [
        ("", ("HEAD", ":WORKTREE:", False)),
        ("HEAD", ("HEAD", ":WORKTREE:", False)),
        ("a", ("a", ":WORKTREE:", True)),
        ("a..", ("a", ":WORKTREE:", False)),
        ("a...", ("a", ":WORKTREE:", True)),
        ("..HEAD", ("HEAD", "HEAD", False)),
        ("...HEAD", ("HEAD", "HEAD", True)),
        ("a..HEAD", ("a", "HEAD", False)),
        ("a...HEAD", ("a", "HEAD", True)),
        ("a..b", ("a", "b", False)),
        ("a...b", ("a", "b", True)),
    ],
)
def test_revisionrange_parse(revision_range, expect):
    """Test for :meth:`RevisionRange.parse`"""
    result = git.RevisionRange._parse(revision_range)

    assert result == expect


def git_call(cmd, encoding=None):
    """Returns a mocked call to git"""
    return call(
        cmd.split(),
        cwd=str(Path("/path")),
        encoding=encoding,
        stderr=PIPE,
        env={"LC_ALL": "C", "PATH": os.environ["PATH"]},
    )


@pytest.mark.kwparametrize(
    dict(
        revision=":WORKTREE:",
        expect_textdocument_calls=[call.from_file(Path("/path/my.txt"))],
    ),
    dict(
        revision="HEAD",
        expect_git_calls=[
            git_call("git show HEAD:./my.txt"),
            git_call("git log -1 --format=%ct HEAD -- my.txt", encoding="utf-8"),
        ],
        expect_textdocument_calls=[
            call.from_bytes(b"1627107028", mtime="2021-07-24 06:10:28.000000 +0000")
        ],
    ),
    dict(
        revision="HEAD^",
        expect_git_calls=[
            git_call("git show HEAD^:./my.txt"),
            git_call("git log -1 --format=%ct HEAD^ -- my.txt", encoding="utf-8"),
        ],
        expect_textdocument_calls=[
            call.from_bytes(b"1627107028", mtime="2021-07-24 06:10:28.000000 +0000")
        ],
    ),
    dict(
        revision="master",
        expect_git_calls=[
            git_call("git show master:./my.txt"),
            git_call("git log -1 --format=%ct master -- my.txt", encoding="utf-8"),
        ],
        expect_textdocument_calls=[
            call.from_bytes(b"1627107028", mtime="2021-07-24 06:10:28.000000 +0000")
        ],
    ),
    expect_git_calls=[],
)
def test_git_get_content_at_revision_obtain_file_content(
    revision, expect_git_calls, expect_textdocument_calls
):
    """``git_get_content_at_revision`` calls Git or reads files based on revision"""
    with patch("darker.git.check_output") as check_output, patch(
        "darker.git.TextDocument"
    ) as text_document_class:
        # this dummy value acts both as a dummy Unix timestamp for the file as well as
        # the contents of the file:
        check_output.return_value = b"1627107028"

        git.git_get_content_at_revision(Path("my.txt"), revision, Path("/path"))

        assert check_output.call_args_list == expect_git_calls
        assert text_document_class.method_calls == expect_textdocument_calls


@pytest.mark.kwparametrize(
    dict(revrange="HEAD", expect="HEAD..:WORKTREE:"),
    dict(revrange="{initial}", expect="{initial}..:WORKTREE:"),
    dict(revrange="{initial}..", expect="{initial}..:WORKTREE:"),
    dict(revrange="{initial}..HEAD", expect="{initial}..HEAD"),
    dict(revrange="{initial}..feature", expect="{initial}..feature"),
    dict(revrange="{initial}...", expect="{initial}..:WORKTREE:"),
    dict(revrange="{initial}...HEAD", expect="{initial}..HEAD"),
    dict(revrange="{initial}...feature", expect="{initial}..feature"),
    dict(revrange="master", expect="{initial}..:WORKTREE:"),
    dict(revrange="master..", expect="master..:WORKTREE:"),
    dict(revrange="master..HEAD", expect="master..HEAD"),
    dict(revrange="master..feature", expect="master..feature"),
    dict(revrange="master...", expect="{initial}..:WORKTREE:"),
    dict(revrange="master...HEAD", expect="{initial}..HEAD"),
    dict(revrange="master...feature", expect="{initial}..feature"),
)
def test_revisionrange_parse_with_common_ancestor(git_repo, revrange, expect):
    """``_git_get_old_revision()`` gets common ancestor using Git when necessary"""
    git_repo.add({"a": "i"}, commit="Initial commit")
    initial = git_repo.get_hash()
    git_repo.add({"a": "m"}, commit="in master")
    master = git_repo.get_hash()
    git_repo.create_branch("feature", initial)
    git_repo.add({"a": "f"}, commit="in feature")

    result = git.RevisionRange.parse_with_common_ancestor(
        revrange.format(initial=initial), git_repo.root
    )

    rev1, rev2 = expect.format(initial=initial, master=master).split("..")
    assert result.rev1 == rev1
    assert result.rev2 == rev2


@pytest.mark.kwparametrize(
    dict(path="file.py", expect="file.py"),
    dict(path="subdir/file.py", expect="subdir/file.py"),
    dict(path="file.py.12345.tmp", expect="file.py"),
    dict(path="subdir/file.py.12345.tmp", expect="subdir/file.py"),
    dict(path="file.py.tmp", expect="file.py.tmp"),
    dict(path="subdir/file.py.tmp", expect="subdir/file.py.tmp"),
    dict(path="file.12345.tmp", expect="file.12345.tmp"),
    dict(path="subdir/file.12345.tmp", expect="subdir/file.12345.tmp"),
)
def test_get_path_in_repo(path, expect):
    """``get_path_in_repo`` drops two suffixes from ``.py.<HASH>.tmp``"""
    result = git.get_path_in_repo(Path(path))

    assert result == Path(expect)


@pytest.mark.kwparametrize(
    dict(path=".", create=False, expect=False),
    dict(path="main", create=True, expect=False),
    dict(path="main.c", create=True, expect=False),
    dict(path="main.py", create=True, expect=True),
    dict(path="main.py", create=False, expect=False),
    dict(path="main.pyx", create=True, expect=False),
    dict(path="main.pyi", create=True, expect=False),
    dict(path="main.pyc", create=True, expect=False),
    dict(path="main.pyo", create=True, expect=False),
    dict(path="main.js", create=True, expect=False),
)
def test_should_reformat_file(tmpdir, path, create, expect):
    if create:
        (tmpdir / path).ensure()

    result = git.should_reformat_file(Path(tmpdir / path))

    assert result == expect


@pytest.mark.kwparametrize(
    dict(cmd=[], exit_on_error=True, expect_template=CalledProcessError(1, "")),
    dict(
        cmd=["status", "-sb"],
        exit_on_error=True,
        expect_template=[
            "## branch",
            "A  add_index.py",
            "D  del_index.py",
            " D del_worktree.py",
            "A  mod_index.py",
            "?? add_worktree.py",
            "?? mod_worktree.py",
        ],
    ),
    dict(
        cmd=["diff"],
        exit_on_error=True,
        expect_template=[
            "diff --git a/del_worktree.py b/del_worktree.py",
            "deleted file mode 100644",
            "index 94f3610..0000000",
            "--- a/del_worktree.py",
            "+++ /dev/null",
            "@@ -1 +0,0 @@",
            "-original",
            "\\ No newline at end of file",
        ],
    ),
    dict(
        cmd=["merge-base", "master"],
        exit_on_error=True,
        expect_template=CalledProcessError(129, ""),
    ),
    dict(
        cmd=["merge-base", "master", "HEAD"],
        exit_on_error=True,
        expect_template=["<hash of branch point>"],
    ),
    dict(
        cmd=["show", "missing.file"],
        exit_on_error=True,
        expect_template=SystemExit(123),
    ),
    dict(
        cmd=["show", "missing.file"],
        exit_on_error=False,
        expect_template=CalledProcessError(128, ""),
    ),
)
def test_git_check_output_lines(branched_repo, cmd, exit_on_error, expect_template):
    """Unit test for :func:`_git_check_output_lines`"""
    if isinstance(expect_template, BaseException):
        expect: Union[List[str], BaseException] = expect_template
    else:
        replacements = {"<hash of branch point>": branched_repo.get_hash("master^")}
        expect = [replacements.get(line, line) for line in expect_template]
    with raises_or_matches(expect, ["returncode", "code"]) as check:

        check(git._git_check_output_lines(cmd, branched_repo.root, exit_on_error))


@pytest.mark.kwparametrize(
    dict(
        cmd=["show", "{initial}:/.file2"],
        exit_on_error=True,
        expect_exc=SystemExit,
        expect_log=(
            r"ERROR    darker\.git:git\.py:\d+ fatal: "
            r"[pP]ath '/\.file2' does not exist in '{initial}'\n$"
        ),
    ),
    dict(
        cmd=["show", "{initial}:/.file2"],
        exit_on_error=False,
        expect_exc=CalledProcessError,
    ),
    dict(
        cmd=["non-existing", "command"],
        exit_on_error=True,
        expect_exc=CalledProcessError,
        expect_stderr="git: 'non-existing' is not a git command. See 'git --help'.\n",
    ),
    dict(
        cmd=["non-existing", "command"],
        exit_on_error=False,
        expect_exc=CalledProcessError,
    ),
    expect_stderr="",
    expect_log=r"$",
)
def test_git_check_output_lines_stderr_and_log(
    git_repo, capfd, caplog, cmd, exit_on_error, expect_exc, expect_stderr, expect_log
):
    """Git non-existing file error is logged and suppressed from stderr"""
    git_repo.add({"file1": "file1"}, commit="Initial commit")
    initial = git_repo.get_hash()[:7]
    git_repo.add({"file2": "file2"}, commit="Second commit")
    capfd.readouterr()  # flush captured stdout and stderr
    cmdline = [s.format(initial=initial) for s in cmd]
    with pytest.raises(expect_exc):

        git._git_check_output_lines(cmdline, git_repo.root, exit_on_error)

    outerr = capfd.readouterr()
    assert outerr.out == ""
    assert outerr.err == expect_stderr
    expect_log_re = expect_log.format(initial=initial)
    assert re.match(expect_log_re, caplog.text), repr(caplog.text)


def test_git_get_content_at_revision_stderr(git_repo, capfd, caplog):
    """No stderr or log output from ``git_get_content_at_revision`` for missing file"""
    git_repo.add({"file1": "file1"}, commit="Initial commit")
    initial = git_repo.get_hash()[:7]
    git_repo.add({"file2": "file2"}, commit="Second commit")
    capfd.readouterr()  # flush captured stdout and stderr

    result = git.git_get_content_at_revision(Path("file2"), initial, git_repo.root)

    assert result == TextDocument()
    outerr = capfd.readouterr()
    assert outerr.out == ""
    assert outerr.err == ""
    assert caplog.text == ""


@pytest.fixture(scope="module")
def encodings_repo(tmp_path_factory):
    """Create an example Git repository using various encodings for the same file"""
    tmpdir = tmp_path_factory.mktemp("branched_repo")
    git_repo = GitRepoFixture.create_repository(tmpdir)
    # Commit without an encoding cookie, defaults to utf-8
    git_repo.add({"file.py": "darker = 'plus foncé'\n"}, commit="Default encoding")
    git_repo.create_tag("default")
    # Commit without an encoding cookie but with a utf-8 signature
    content = "darker = 'plus foncé'\n".encode("utf-8-sig")
    git_repo.add({"file.py": content}, commit="utf-8-sig")
    git_repo.create_tag("utf-8-sig")
    # Commit with an iso-8859-1 encoding cookie
    content = "# coding: iso-8859-1\ndarker = 'plus foncé'\n".encode("iso-8859-1")
    git_repo.add({"file.py": content}, commit="iso-8859-1")
    git_repo.create_tag("iso-8859-1")
    # Commit with a utf-8 encoding cookie
    content = "# coding: utf-8\npython = 'パイソン'\n".encode()
    git_repo.add({"file.py": content}, commit="utf-8")
    git_repo.create_tag("utf-8")
    # Current worktree content (not committed) with a shitfjs encoding cookie
    content = "# coding: shiftjis\npython = 'パイソン'\n".encode("shiftjis")
    git_repo.add({"file.py": content})
    return git_repo


@pytest.mark.kwparametrize(
    dict(commit="default", encoding="utf-8", lines=("darker = 'plus foncé'",)),
    dict(commit="utf-8-sig", encoding="utf-8-sig", lines=("darker = 'plus foncé'",)),
    dict(
        commit="iso-8859-1",
        encoding="iso-8859-1",
        lines=("# coding: iso-8859-1", "darker = 'plus foncé'"),
    ),
    dict(
        commit="utf-8", encoding="utf-8", lines=("# coding: utf-8", "python = 'パイソン'")
    ),
    dict(
        commit=":WORKTREE:",
        encoding="shiftjis",
        lines=("# coding: shiftjis", "python = 'パイソン'"),
    ),
)
def test_git_get_content_at_revision_encoding(encodings_repo, commit, encoding, lines):
    """Git file is loaded using its historical encoding"""
    result = git.git_get_content_at_revision(
        Path("file.py"), commit, encodings_repo.root
    )
    assert result.encoding == encoding
    assert result.lines == lines


@pytest.mark.kwparametrize(
    dict(retval=0, expect=True),
    dict(retval=1, expect=False),
    dict(retval=2, expect=False),
)
def test_git_exists_in_revision_git_call(retval, expect):
    """``_git_exists_in_revision()`` calls Git and converts return value correctly"""
    with patch.object(git, "run") as run:
        run.return_value.returncode = retval

        result = git._git_exists_in_revision(Path("path.py"), "rev2", Path("."))

    run.assert_called_once_with(
        ["git", "cat-file", "-e", "rev2:path.py"],
        cwd=".",
        check=False,
        stderr=DEVNULL,
        env={"LC_ALL": "C", "PATH": os.environ["PATH"]},
    )
    assert result == expect


@pytest.mark.kwparametrize(
    dict(rev2="{add}", path="dir/a.py", expect=True),
    dict(rev2="{add}", path="dir/b.py", expect=True),
    dict(rev2="{add}", path="dir/", expect=True),
    dict(rev2="{add}", path="dir", expect=True),
    dict(rev2="{del_a}", path="dir/a.py", expect=False),
    dict(rev2="{del_a}", path="dir/b.py", expect=True),
    dict(rev2="{del_a}", path="dir/", expect=True),
    dict(rev2="{del_a}", path="dir", expect=True),
    dict(rev2="HEAD", path="dir/a.py", expect=False),
    dict(rev2="HEAD", path="dir/b.py", expect=False),
    dict(rev2="HEAD", path="dir/", expect=False),
    dict(rev2="HEAD", path="dir", expect=False),
)
def test_git_exists_in_revision(git_repo, rev2, path, expect):
    """``_get_exists_in_revision()`` detects file/dir existence correctly"""
    git_repo.add({"dir/a.py": "", "dir/b.py": ""}, commit="Add dir/*.py")
    add = git_repo.get_hash()
    git_repo.add({"dir/a.py": None}, commit="Delete dir/a.py")
    del_a = git_repo.get_hash()
    git_repo.add({"dir/b.py": None}, commit="Delete dir/b.py")

    result = git._git_exists_in_revision(
        Path(path), rev2.format(add=add, del_a=del_a), git_repo.root
    )

    assert result == expect


@pytest.mark.kwparametrize(
    dict(rev2="{add}", expect=set()),
    dict(rev2="{del_a}", expect={Path("dir/a.py")}),
    dict(rev2="HEAD", expect={Path("dir"), Path("dir/a.py"), Path("dir/b.py")}),
)
def test_get_missing_at_revision(git_repo, rev2, expect):
    """``get_missing_at_revision()`` returns missing files/directories correctly"""
    git_repo.add({"dir/a.py": "", "dir/b.py": ""}, commit="Add dir/*.py")
    add = git_repo.get_hash()
    git_repo.add({"dir/a.py": None}, commit="Delete dir/a.py")
    del_a = git_repo.get_hash()
    git_repo.add({"dir/b.py": None}, commit="Delete dir/b.py")

    result = git.get_missing_at_revision(
        {Path("dir"), Path("dir/a.py"), Path("dir/b.py")},
        rev2.format(add=add, del_a=del_a),
        git_repo.root,
    )

    assert result == expect


def test_get_missing_at_revision_worktree(git_repo):
    """``get_missing_at_revision()`` returns missing work tree files/dirs correctly"""
    paths = git_repo.add({"dir/a.py": "", "dir/b.py": ""}, commit="Add dir/*.py")
    paths["dir/a.py"].unlink()
    paths["dir/b.py"].unlink()

    result = git.get_missing_at_revision(
        {Path("dir"), Path("dir/a.py"), Path("dir/b.py")}, git.WORKTREE, git_repo.root
    )

    assert result == {Path("dir/a.py"), Path("dir/b.py")}


def test_git_diff_name_only(git_repo):
    """``_git_diff_name_only()`` only returns paths of modified files"""
    git_repo.add({"a.py": "a", "b.py": "b", "c.py": "c"}, commit="Initial commit")
    first = git_repo.get_hash()
    git_repo.add({"a.py": "A", "b.dy": "B"}, commit="only a.py modified")
    second = git_repo.get_hash()

    result = git._git_diff_name_only(
        first, second, {Path("a.py"), Path("c.py"), Path("Z.py")}, git_repo.root
    )

    assert result == {Path("a.py")}


def test_git_ls_files_others(git_repo):
    """``_git_ls_files_others()`` only returns paths of untracked non-ignored files"""
    git_repo.add(
        {
            "tracked.py": "tracked",
            "tracked.ignored": "tracked",
            ".gitignore": "*.ignored",
        },
        commit="Initial commit",
    )
    (git_repo.root / "untracked.py").write_text("untracked")
    (git_repo.root / "untracked.ignored").write_text("untracked")

    result = git._git_ls_files_others(
        {
            Path("tracked.py"),
            Path("tracked.ignored"),
            Path("untracked.py"),
            Path("untracked.ignored"),
            Path("missing.py"),
            Path("missing.ignored"),
        },
        git_repo.root,
    )

    assert result == {Path("untracked.py")}


@pytest.mark.kwparametrize(
    dict(paths=["a.py"], expect=[]),
    dict(expect=[]),
    dict(modify_paths={"a.py": "new"}, expect=["a.py"]),
    dict(modify_paths={"a.py": "new"}, paths=["b.py"], expect=[]),
    dict(modify_paths={"a.py": "new"}, paths=["a.py", "b.py"], expect=["a.py"]),
    dict(
        modify_paths={"c/d.py": "new"}, paths=["c/d.py", "d/f/g.py"], expect=["c/d.py"]
    ),
    dict(modify_paths={"c/e.js": "new"}, paths=["c/e.js"], expect=[]),
    dict(modify_paths={"a.py": "original"}, paths=["a.py"], expect=[]),
    dict(modify_paths={"a.py": None}, paths=["a.py"], expect=[]),
    dict(modify_paths={"h.py": "untracked"}, paths=["h.py"], expect=["h.py"]),
    dict(paths=["h.py"], expect=[]),
    modify_paths={},
    paths=[],
)
def test_git_get_modified_python_files(git_repo, modify_paths, paths, expect):
    """Tests for `darker.git.git_get_modified_python_files()`"""
    root = Path(git_repo.root)
    git_repo.add(
        {
            "a.py": "original",
            "b.py": "original",
            "c/d.py": "original",
            "c/e.js": "original",
            "d/f/g.py": "original",
        },
        commit="Initial commit",
    )
    for path, content in modify_paths.items():
        absolute_path = git_repo.root / path
        if content is None:
            absolute_path.unlink()
        else:
            absolute_path.parent.mkdir(parents=True, exist_ok=True)
            absolute_path.write_bytes(content.encode("ascii"))
    revrange = git.RevisionRange("HEAD", ":WORKTREE:")

    result = git.git_get_modified_python_files(
        {root / p for p in paths}, revrange, cwd=root
    )

    assert result == {Path(p) for p in expect}


@pytest.fixture(scope="module")
def branched_repo(tmp_path_factory):
    """Create an example Git repository with a master branch and a feature branch

    The history created is::

        . worktree
        . index
        * branch
        | * master
        |/
        * Initial commit

    """
    tmpdir = tmp_path_factory.mktemp("branched_repo")
    git_repo = GitRepoFixture.create_repository(tmpdir)
    git_repo.add(
        {
            "del_master.py": "original",
            "del_branch.py": "original",
            "del_index.py": "original",
            "del_worktree.py": "original",
            "mod_master.py": "original",
            "mod_branch.py": "original",
            "mod_both.py": "original",
            "mod_same.py": "original",
            "keep.py": "original",
        },
        commit="Initial commit",
    )
    branch_point = git_repo.get_hash()
    git_repo.add(
        {
            "del_master.py": None,
            "add_master.py": "master",
            "mod_master.py": "master",
            "mod_both.py": "master",
            "mod_same.py": "same",
        },
        commit="master",
    )
    git_repo.create_branch("branch", branch_point)
    git_repo.add(
        {
            "del_branch.py": None,
            "mod_branch.py": "branch",
            "mod_both.py": "branch",
            "mod_same.py": "same",
        },
        commit="branch",
    )
    git_repo.add(
        {"del_index.py": None, "add_index.py": "index", "mod_index.py": "index"}
    )
    (git_repo.root / "del_worktree.py").unlink()
    (git_repo.root / "add_worktree.py").write_bytes(b"worktree")
    (git_repo.root / "mod_worktree.py").write_bytes(b"worktree")
    return git_repo


@pytest.mark.kwparametrize(
    dict(
        _description="from latest commit in branch to worktree and index",
        revrange="HEAD",
        expect={"add_index.py", "add_worktree.py", "mod_index.py", "mod_worktree.py"},
    ),
    dict(
        _description="from initial commit to worktree and index on branch (implicit)",
        revrange="master",
        expect={
            "mod_both.py",
            "mod_same.py",
            "mod_branch.py",
            "add_index.py",
            "mod_index.py",
            "add_worktree.py",
            "mod_worktree.py",
        },
    ),
    dict(
        _description="from initial commit to worktree and index on branch",
        revrange="master...",
        expect={
            "mod_both.py",
            "mod_same.py",
            "mod_branch.py",
            "add_index.py",
            "mod_index.py",
            "add_worktree.py",
            "mod_worktree.py",
        },
    ),
    dict(
        _description="from master to worktree and index on branch",
        revrange="master..",
        expect={
            "del_master.py",
            "mod_master.py",
            "mod_both.py",
            "mod_branch.py",
            "add_index.py",
            "mod_index.py",
            "add_worktree.py",
            "mod_worktree.py",
        },
    ),
    dict(
        _description=(
            "from master to last commit on branch," " excluding worktree and index"
        ),
        revrange="master..HEAD",
        expect={
            "del_master.py",
            "mod_master.py",
            "mod_both.py",
            "mod_branch.py",
        },
    ),
    dict(
        _description="from master to branch, excluding worktree and index",
        revrange="master..branch",
        expect={
            "del_master.py",
            "mod_master.py",
            "mod_both.py",
            "mod_branch.py",
        },
    ),
    dict(
        _description=(
            "from initial commit to last commit on branch,"
            " excluding worktree and index"
        ),
        revrange="master...HEAD",
        expect={"mod_both.py", "mod_same.py", "mod_branch.py"},
    ),
    dict(
        _description="from initial commit to previous commit on branch",
        revrange="master...branch",
        expect={"mod_both.py", "mod_same.py", "mod_branch.py"},
    ),
)
def test_git_get_modified_python_files_revision_range(
    _description, branched_repo, revrange, expect
):
    """Test for :func:`darker.git.git_get_modified_python_files` with revision range"""
    result = git.git_get_modified_python_files(
        [Path(branched_repo.root)],
        git.RevisionRange.parse_with_common_ancestor(revrange, branched_repo.root),
        Path(branched_repo.root),
    )

    assert {path.name for path in result} == expect


@pytest.mark.kwparametrize(
    dict(
        environ={},
        expect_rev1="HEAD",
        expect_rev2=":WORKTREE:",
        expect_use_common_ancestor=False,
    ),
    dict(
        environ={"PRE_COMMIT_FROM_REF": "old"},
        expect_rev1="HEAD",
        expect_rev2=":WORKTREE:",
        expect_use_common_ancestor=False,
    ),
    dict(
        environ={"PRE_COMMIT_TO_REF": "new"},
        expect_rev1="HEAD",
        expect_rev2=":WORKTREE:",
        expect_use_common_ancestor=False,
    ),
    dict(
        environ={"PRE_COMMIT_FROM_REF": "old", "PRE_COMMIT_TO_REF": "new"},
        expect_rev1="old",
        expect_rev2="new",
        expect_use_common_ancestor=True,
    ),
)
def test_revisionrange_parse_pre_commit(
    environ, expect_rev1, expect_rev2, expect_use_common_ancestor
):
    """RevisionRange._parse(':PRE-COMMIT:') gets the range from environment variables"""
    with patch.dict(os.environ, environ):

        result = git.RevisionRange._parse(":PRE-COMMIT:")

        assert result == (expect_rev1, expect_rev2, expect_use_common_ancestor)


edited_linenums_differ_cases = pytest.mark.kwparametrize(
    dict(context_lines=0, expect=[3, 7]),
    dict(context_lines=1, expect=[2, 3, 4, 6, 7, 8]),
    dict(context_lines=2, expect=[1, 2, 3, 4, 5, 6, 7, 8]),
    dict(context_lines=3, expect=[1, 2, 3, 4, 5, 6, 7, 8]),
)


@edited_linenums_differ_cases
def test_edited_linenums_differ_compare_revisions(git_repo, context_lines, expect):
    """Tests for EditedLinenumsDiffer.revision_vs_worktree()"""
    paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
    paths["a.py"].write_bytes(b"1\n2\nthree\n4\n5\n6\nseven\n8\n")
    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
    differ = git.EditedLinenumsDiffer(git_repo.root, revrange)

    linenums = differ.compare_revisions(Path("a.py"), context_lines)

    assert linenums == expect


@edited_linenums_differ_cases
def test_edited_linenums_differ_revision_vs_lines(git_repo, context_lines, expect):
    """Tests for EditedLinenumsDiffer.revision_vs_lines()"""
    git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
    content = TextDocument.from_lines(["1", "2", "three", "4", "5", "6", "seven", "8"])
    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
    differ = git.EditedLinenumsDiffer(git_repo.root, revrange)

    linenums = differ.revision_vs_lines(Path("a.py"), content, context_lines)

    assert linenums == expect


@pytest.mark.kwparametrize(
    dict(context_lines=0, expect=[1, 3, 4, 5, 6, 8]),
    dict(context_lines=1, expect=[1, 2, 3, 4, 5, 6, 7, 8]),
)
def test_edited_linenums_differ_revision_vs_lines_multiline_strings(
    git_repo, context_lines, expect
):
    """Tests for EditedLinenumsDiffer.revision_vs_lines() with multi-line strings"""
    git_repo.add(
        {
            "a.py": dedent(
                """\
                change\n
                keep\n
                '''change first,\n
                keep second\n
                and third,\n
                change fourth line of multiline'''\n
                keep\n
                change\n
                """
            )
        },
        commit="Initial commit",
    )
    content = TextDocument.from_lines(
        [
            "CHANGED",
            "keep",
            "'''CHANGED FIRST,",
            "keep second",
            "and third,",
            "CHANGED FOURTH LINE OF MULTILINE'''",
            "keep",
            "CHANGED",
        ]
    )
    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
    differ = git.EditedLinenumsDiffer(git_repo.root, revrange)

    linenums = differ.revision_vs_lines(Path("a.py"), content, context_lines)

    assert linenums == expect


def test_local_gitconfig_ignored_by_gitrepofixture(tmp_path):
    """Tests that ~/.gitconfig is ignored when running darker's git tests"""
    (tmp_path / "HEAD").write_text("ref: refs/heads/main")

    # Note: once we decide to drop support for git < 2.28, the HEAD file
    # creation above can be removed, and setup can simplify to
    # check_call("git config --global init.defaultBranch main".split())
    check_call(  # nosec
        "git config --global init.templateDir".split() + [str(tmp_path)],
        env={"HOME": str(tmp_path), "PATH": os.environ["PATH"]},
    )
    root = tmp_path / "repo"
    root.mkdir()
    git_repo = GitRepoFixture.create_repository(root)
    assert git_repo.get_branch() == "master"



"""Custom formatter and action for argparse"""

import logging
import re
import sys
from argparse import SUPPRESS, Action, ArgumentParser, HelpFormatter, Namespace
from textwrap import fill
from typing import Any, List, Optional, Sequence, Union

WORD_RE = re.compile(r"\w")


def _fill_line(line: str, width: int, indent: str) -> str:
    first_word_match = WORD_RE.search(line)
    first_word_offset = first_word_match.start() if first_word_match else 0
    return fill(
        line,
        width,
        initial_indent=indent,
        subsequent_indent=indent + first_word_offset * " ",
    )


class NewlinePreservingFormatter(HelpFormatter):
    def _fill_text(self, text: str, width: int, indent: str) -> str:
        if "\n" in text:
            return "\n".join(
                _fill_line(line, width, indent) for line in text.split("\n")
            )
        return super()._fill_text(text, width, indent)


class OptionsForReadmeAction(Action):
    """Implementation of the ``--options-for-readme`` argument

    This argparse action prints optional command line arguments in a format suitable for
    inclusion in ``README.rst``.

    """

    # pylint: disable=too-few-public-methods

    def __init__(
        self, option_strings: List[str], dest: str = SUPPRESS, help: str = None
    ):  # pylint: disable=redefined-builtin
        super().__init__(option_strings, dest, 0)

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Optional[Union[str, Sequence[Any]]],
        option_string: str = None,
    ) -> None:
        optional_arguments_group = next(
            group
            for group in parser._action_groups
            # The group title for options differs between Python versions
            if group.title in {"optional arguments", "options"}
        )
        actions = []
        for action in optional_arguments_group._group_actions:
            if action.dest in {"help", "version", "options_for_readme"}:
                continue
            if action.help is not None:
                action.help = action.help.replace("`", "``")
            actions.append(action)
        formatter = HelpFormatter(parser.prog, max_help_position=7, width=88)
        formatter.add_arguments(actions)
        sys.stderr.write(formatter.format_help())
        parser.exit()


class LogLevelAction(Action):  # pylint: disable=too-few-public-methods
    """Support for command line actions which increment/decrement the log level"""

    def __init__(  # pylint: disable=too-many-arguments
        self,
        option_strings: List[str],
        dest: str,
        const: int,
        default: int = logging.WARNING,
        required: bool = False,
        help: str = None,  # pylint: disable=redefined-builtin
        metavar: str = None,
    ):
        super().__init__(
            option_strings=option_strings,
            dest=dest,
            nargs=0,
            const=const,
            default=default,
            required=required,
            help=help,
            metavar=metavar,
        )

    def __call__(
        self,
        parser: ArgumentParser,
        namespace: Namespace,
        values: Union[str, Sequence[Any], None],
        option_string: str = None,
    ) -> None:
        current_level = getattr(namespace, self.dest, self.default)
        new_level = current_level + self.const
        new_level = max(new_level, logging.DEBUG)
        new_level = min(new_level, logging.CRITICAL)
        setattr(namespace, self.dest, new_level)



"""Unit tests for :mod:`darker.verification`"""

from typing import List

import pytest

from darker.utils import DiffChunk, TextDocument
from darker.verification import (
    ASTVerifier,
    BinarySearch,
    NotEquivalentError,
    verify_ast_unchanged,
)


@pytest.mark.kwparametrize(
    dict(dst_content=["if False: pass"], expect=AssertionError),
    dict(dst_content=["if True:", "    pass"], expect=None),
)
def test_verify_ast_unchanged(dst_content, expect):
    black_chunks: List[DiffChunk] = [(1, ("black",), ("chunks",))]
    edited_linenums = [1, 2]
    try:
        verify_ast_unchanged(
            TextDocument.from_lines(["if True: pass"]),
            TextDocument.from_lines(dst_content),
            black_chunks,
            edited_linenums,
        )
    except NotEquivalentError:
        assert expect is AssertionError
    else:
        assert expect is None


def test_ast_verifier_is_equivalent():
    """``darker.verification.ASTVerifier.is_equivalent_to_baseline``"""
    verifier = ASTVerifier(baseline=TextDocument.from_lines(["if True: pass"]))
    assert verifier.is_equivalent_to_baseline(
        TextDocument.from_lines(["if True:", "    pass"])
    )
    assert not verifier.is_equivalent_to_baseline(
        TextDocument.from_lines(["if False: pass"])
    )
    assert not verifier.is_equivalent_to_baseline(
        TextDocument.from_lines(["if False:"])
    )


def test_binary_search_premature_result():
    """``darker.verification.BinarySearch``"""
    with pytest.raises(RuntimeError):

        _ = BinarySearch(0, 5).result


def test_binary_search():
    """``darker.verification.BinarySearch``"""
    search = BinarySearch(0, 5)
    tries = []
    while not search.found:
        tries.append(search.get_next())

        search.respond(tries[-1] > 2)
    assert search.result == 3
    assert tries == [0, 3, 2]


@pytest.mark.parametrize("i", range(50))
def test_binary_search_in_50(i):
    """Simple 'fuzzy test' for BinarySearch"""
    search = BinarySearch(0, 50)
    while not search.found:
        search.respond(search.get_next() >= i)
    assert search.result == i



"""Highlighting of terminal output"""

# pylint: disable=import-outside-toplevel,unused-import

import sys
from typing import Optional, cast


def should_use_color(config_color: Optional[bool]) -> bool:
    """Return ``True`` if configuration and package support allow output highlighting

    In ``config_color``, the combination of ``color =`` in ``pyproject.toml``, the
    ``PY_COLORS`` environment variable, and the ``--color``/``--no-color`` command line
    options determine whether the user wants to force enable or disable highlighting.

    If highlighting isn't forced either way, it is automatically enabled for terminal
    output.

    Finally, if ``pygments`` isn't installed, highlighting is disabled.

    :param config_color: The configuration as parsed from ``pyproject.toml`` and
                         overridden using environment variables and/or command line
                         options
    :return: ``True`` if highlighting should be used

    """
    if config_color is not None:
        use_color = config_color
    else:
        use_color = sys.stdout.isatty()

    if use_color:
        try:
            import pygments  # noqa

            return True
        except ImportError:
            pass
    return False


def colorize(output: str, lexer_name: str, use_color: bool) -> str:
    """Return the output highlighted for terminal if Pygments is available"""
    if not use_color:
        return output
    from pygments import highlight
    from pygments.formatters.terminal import TerminalFormatter
    from pygments.lexers import get_lexer_by_name

    lexer = get_lexer_by_name(lexer_name)
    highlighted = highlight(output, lexer, TerminalFormatter())
    if "\n" not in output:
        # see https://github.com/pygments/pygments/issues/1107
        highlighted = highlighted.rstrip("\n")
    return cast(str, highlighted)



import logging
import sys
from pathlib import Path
from typing import Any, Collection, List, Optional

from darker.black_compat import find_project_root
from darker.diff import diff_chunks
from darker.exceptions import IncompatiblePackageError, MissingPackageError
from darker.git import EditedLinenumsDiffer
from darker.utils import DiffChunk, TextDocument, glob_any

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


try:
    import isort

    # Work around Mypy problem
    # https://github.com/python/mypy/issues/7030#issuecomment-504128883
    try:
        isort_code = getattr(isort, "code")  # noqa: B009
    except AttributeError:
        # Postpone error message about incompatbile `isort` version until `--isort` is
        # actually used.
        def isort_code(*args: Any, **kwargs: Any) -> str:  # type: ignore[misc]
            """Fake `isort.code()` function to use when `isort < 5` is installed"""
            raise IncompatiblePackageError(
                "An incompatible 'isort' package was found. Please install version"
                " 5.0.0 or later."
            )
except ImportError:
    # `isort` is an optional dependency. Prevent the `ImportError` if it's missing.
    isort = None  # type: ignore[assignment]

    def isort_code(*args: Any, **kwargs: Any) -> str:  # type: ignore[misc]
        """Fake `isort.code()` function to use when `isort` isn't installed"""
        raise MissingPackageError(
            "No module named 'isort'. Please install the 'isort' package before using"
            " the --isort / -i option."
        )


__all__ = ["apply_isort", "isort"]

logger = logging.getLogger(__name__)


class IsortArgs(TypedDict, total=False):
    line_length: int
    settings_file: str
    settings_path: str


def apply_isort(
    content: TextDocument,
    src: Path,
    exclude: Collection[str],
    edited_linenums_differ: EditedLinenumsDiffer,
    config: Optional[str] = None,
    line_length: Optional[int] = None,
) -> TextDocument:
    """Run isort on the given Python source file content

    :param content: The contents of the Python source code file to sort imports in
    :param src: The relative path to the file. This must be the actual path in the
                repository, which may differ from the path given on the command line in
                case of VSCode temporary files.
    :param exclude: The file path patterns to exclude from import sorting
    :param edited_linenums_differ: Helper for finding out which lines were edited
    :param config: Path to configuration file
    :param line_length: Maximum line length to use

    """
    if glob_any(src, exclude):
        return content
    edited_linenums = edited_linenums_differ.revision_vs_lines(
        src,
        content,
        context_lines=0,
    )
    if not edited_linenums:
        return content
    isort_args = _build_isort_args(src, config, line_length)
    rev2_isorted = _call_isort_code(content, isort_args)
    # Get the chunks in the diff between the edited and import-sorted file
    isort_chunks = diff_chunks(content, rev2_isorted)
    if not isort_chunks:
        # No imports were sorted. Return original content.
        return content
    if not _diff_overlaps_with_edits(edited_linenums, isort_chunks):
        # No lines had been modified in the range of modified import lines. Return
        # original content.
        return content
    # The range lines modified by sorted imports overlaps with user modifications in the
    # code. Return the import-sorted file.
    return rev2_isorted


def _build_isort_args(
    src: Path,
    config: Optional[str] = None,
    line_length: Optional[int] = None,
) -> IsortArgs:
    """Build ``isort.code()`` keyword arguments

    :param src: The relative path to the file. This must be the actual path in the
                repository, which may differ from the path given on the command line in
                case of VSCode temporary files.
    :param config: Path to configuration file
    :param line_length: Maximum line length to use

    """
    isort_args: IsortArgs = {}
    if config:
        isort_args["settings_file"] = config
    else:
        isort_args["settings_path"] = str(find_project_root((str(src),)))
    if line_length:
        isort_args["line_length"] = line_length
    return isort_args


def _call_isort_code(content: TextDocument, isort_args: IsortArgs) -> TextDocument:
    """Call ``isort.code()`` and return the result as a `TextDocument` object

    :param content: The contents of the Python source code file to sort imports in
    :param isort_args: Keyword arguments for ``isort.code()``

    """
    code = content.string
    logger.debug(
        "isort.code(code=..., %s)",
        ", ".join(f"{k}={v!r}" for k, v in isort_args.items()),
    )
    try:
        code = isort_code(code=code, **isort_args)
    except isort.exceptions.FileSkipComment:
        pass
    return TextDocument.from_str(
        code,
        encoding=content.encoding,
        mtime=content.mtime,
    )


def _diff_overlaps_with_edits(
    edited_linenums: List[int], isort_chunks: List[DiffChunk]
) -> bool:
    """Return ``True`` if the complete diff overlaps the range of edited lines

    :param edited_linenums: The line numbers of all edited lines
    :param isort_chunks: The diff chunks
    :return: ``True`` if the two overlap

    """
    if not edited_linenums:
        return False
    first_edited_linenum, last_edited_linenum = edited_linenums[0], edited_linenums[-1]
    modified_chunks = [
        (linenum, old, new) for linenum, old, new in isort_chunks if old != new
    ]
    if not modified_chunks:
        return False
    (first_isort_line, _, _) = modified_chunks[0]
    (last_isort_chunk_start, last_isort_chunk_original_lines, _) = modified_chunks[-1]
    last_isort_line = last_isort_chunk_start + len(last_isort_chunk_original_lines)
    return (
        first_edited_linenum < last_isort_line
        and last_edited_linenum >= first_isort_line
    )



from textwrap import dedent

ORIGINAL = dedent(
    '''\
    original first line
    original second line
    original third line
'''
)

CHANGED = dedent(
    '''\
    original first line
    changed second line
    original third line
'''
)

BLANK_SEP_ORIGINAL = dedent(
    """\
    original first line

    original third line

    original fifth line
    """
)

BLANK_SEP_CHANGED = dedent(
    """\
    changed first line

    original third line

    changed fifth line
    """
)

BLANK_SEP_SANDWICH_ORIGINAL = dedent(
    """\
    original first line

    original third line

    original fifth line
    """
)

BLANK_SEP_SANDWICH_CHANGED = dedent(
    """\
    changed first line

    original third line

    changed fifth line
    """
)


BLANK_SEP_ORIGINAL = dedent(
    """\
    original first line

    original third line

    original fifth line
    """
)

BLANK_SEP_CHANGED = dedent(
    """\
    changed first line

    changed third line

    changed fifth line
    """
)



"""Functions for maintaining compatibility with multiple Black versions"""

from pathlib import Path
from typing import Any, Sequence, Tuple, cast

from black import find_project_root as black_find_project_root


def find_project_root(srcs: Sequence[str]) -> Path:
    """Hide changed return value type in Black behind this wrapper

    :param srcs: Files and directories to find the common root for
    :return: Project root path

    """
    root = cast(Any, black_find_project_root(tuple(srcs or ["."])))
    if isinstance(root, tuple):
        # Black >= 22.1
        return cast(Tuple[Path], root)[0]
    # Black < 22
    return cast(Path, root)



from difflib import SequenceMatcher

from darker.tests.git_diff_example_output import (
    BLANK_SEP_CHANGED,
    BLANK_SEP_ORIGINAL,
    BLANK_SEP_SANDWICH_CHANGED,
    BLANK_SEP_SANDWICH_ORIGINAL,
    CHANGED,
    ORIGINAL,
)


def test_sequencematcher():
    """``SequenceMatcher`` detects a single changed line in between correctly"""
    matcher = SequenceMatcher(
        None, ORIGINAL.splitlines(), CHANGED.splitlines(), autojunk=False
    )
    assert matcher.get_opcodes() == [
        ("equal", 0, 1, 0, 1),
        ("replace", 1, 2, 1, 2),
        ("equal", 2, 3, 2, 3),
    ]


def test_sequencematcher_blank_separated_changes_unchanged_sandwiched():
    """``SequenceMatcher`` detects blank line delimited unchanged region correctly"""
    matcher = SequenceMatcher(
        None,
        BLANK_SEP_SANDWICH_ORIGINAL.splitlines(),
        BLANK_SEP_SANDWICH_CHANGED.splitlines(),
        autojunk=False,
    )
    assert matcher.get_opcodes() == [
        ("replace", 0, 1, 0, 1),
        ("equal", 1, 4, 1, 4),
        ("replace", 4, 5, 4, 5),
    ]


def test_sequencematcher_blank_separated_changes():
    """``SequenceMatcher`` detects blank line delimited changes correctly"""
    matcher = SequenceMatcher(
        None,
        BLANK_SEP_ORIGINAL.splitlines(),
        BLANK_SEP_CHANGED.splitlines(),
        autojunk=False,
    )
    assert matcher.get_opcodes() == [
        ("replace", 0, 1, 0, 1),
        ("equal", 1, 2, 1, 2),
        ("replace", 2, 3, 2, 3),
        ("equal", 3, 4, 3, 4),
        ("replace", 4, 5, 4, 5),
    ]



"""Help and usage instruction texts used for the command line parser"""

ISORT_INSTRUCTION = "Please run `pip install darker[isort]`"

DESCRIPTION_PARTS = [
    "Re-format Python source files by using\n",
    "- `isort` to sort Python import definitions alphabetically within logical"
    " sections\n",
    "- `black` to re-format code changed since the last Git commit",
]
try:
    import isort
except ImportError:
    isort = None  # type: ignore[assignment]
    DESCRIPTION_PARTS.extend(
        ["\n", "\n", f"{ISORT_INSTRUCTION} to enable sorting of import definitions"]
    )
DESCRIPTION = "".join(DESCRIPTION_PARTS)


SRC = "Path(s) to the Python source file(s) to reformat"

REVISION = (
    "Git revision against which to compare the working tree. Tags, branch names, commit"
    " hashes, and other expressions like `HEAD~5` work here. Also a range like"
    " `master...HEAD` or `master...` can be used to compare the best common ancestor."
    " With the magic value `:PRE-COMMIT:`, Darker works in pre-commit compatible mode."
    " Darker expects the revision range from the `PRE_COMMIT_FROM_REF` and"
    " `PRE_COMMIT_TO_REF` environment variables. If those are not found, Darker works"
    " against `HEAD`."
)

DIFF = (
    "Don't write the files back, just output a diff for each file on stdout."
    " Highlight syntax if on a terminal and the `pygments` package is available, or if"
    " enabled by configuration."
)

CHECK = (
    "Don't write the files back, just return the status.  Return code 0 means"
    " nothing would change.  Return code 1 means some files would be"
    " reformatted."
)

STDOUT = (
    "Force complete reformatted output to stdout, instead of in-place. Only valid if"
    " there's just one file to reformat. Highlight syntax if on a terminal and the"
    " `pygments` package is available, or if enabled by configuration."
)

ISORT_PARTS = ["Also sort imports using the `isort` package"]
if not isort:
    ISORT_PARTS.append(f". {ISORT_INSTRUCTION} to enable usage of this option.")
ISORT = "".join(ISORT_PARTS)

LINT = (
    "Also run a linter on changed files. `CMD` can be a name of path of the"
    " linter binary, or a full quoted command line. Highlight linter output syntax if"
    " on a terminal and the `pygments` package is available, or if enabled by"
    " configuration."
)

CONFIG = "Ask `black` and `isort` to read configuration from `PATH`."

VERBOSE = "Show steps taken and summarize modifications"
QUIET = "Reduce amount of output"
COLOR = (
    "Enable syntax highlighting even for non-terminal output. Overrides the environment"
    " variable PY_COLORS=0"
)
NO_COLOR = (
    "Disable syntax highlighting even for terminal output. Overrides the environment"
    " variable PY_COLORS=1"
)

VERSION = "Show the version of `darker`"

SKIP_STRING_NORMALIZATION = "Don't normalize string quotes or prefixes"
NO_SKIP_STRING_NORMALIZATION = (
    "Normalize string quotes or prefixes. This can be used to override"
    " `skip_string_normalization = true` from a configuration file."
)
SKIP_MAGIC_TRAILING_COMMA = (
    "Skip adding trailing commas to expressions that are split by comma"
    " where each element is on its own line. This includes function signatures."
    " This can be used to override"
    " `skip_magic_trailing_comma = true` from a configuration file."
)

LINE_LENGTH = "How many characters per line to allow [default: 88]"

WORKERS = "How many parallel workers to allow, or `0` for one per core [default: 1]"



"""Tests for `darker.concurrency`"""

from concurrent.futures import Future, ProcessPoolExecutor
from unittest.mock import Mock

import pytest

from darker import concurrency


def test_dummy_executor_submit_success():
    """Function is executed synchronously and future gives the return value"""
    executor = concurrency.DummyExecutor()
    func = Mock(return_value=42)

    future = executor.submit(func, "arg", kwarg="kwarg")

    assert isinstance(future, Future)
    assert future.result() == 42
    func.assert_called_once_with("arg", kwarg="kwarg")


def test_dummy_executor_submit_exception():
    """Future raises the exception from the function"""
    executor = concurrency.DummyExecutor()
    func = Mock(side_effect=ValueError("my-exception"))

    future = executor.submit(func, "arg", kwarg="kwarg")

    assert isinstance(future, Future)
    with pytest.raises(ValueError) as exc:
        assert future.result() == 42
        assert str(exc.value) == "my-exception"


@pytest.mark.kwparametrize(
    dict(max_workers=0, expect=ProcessPoolExecutor),
    dict(max_workers=1, expect=concurrency.DummyExecutor),
    dict(max_workers=2, expect=ProcessPoolExecutor),
    dict(max_workers=42, expect=ProcessPoolExecutor),
)
def test_get_executor(max_workers, expect):
    """A correct executor object is returned based on ``max_workers``"""
    result = concurrency.get_executor(max_workers)

    assert isinstance(result, expect)



"""Unit tests for :mod:`darker.__main__`"""

# pylint: disable=unused-argument,too-many-arguments,use-dict-literal,protected-access

import logging
import re
from argparse import ArgumentError
from pathlib import Path
from textwrap import dedent
from types import SimpleNamespace
from unittest.mock import call, patch

import pytest

import darker.__main__
import darker.import_sorting
import darker.linting
from darker.config import Exclusions
from darker.exceptions import MissingPackageError
from darker.git import WORKTREE, EditedLinenumsDiffer, RevisionRange
from darker.tests.helpers import isort_present
from darker.utils import TextDocument, joinlines
from darker.verification import NotEquivalentError


def _replace_diff_timestamps(text, replacement="<timestamp>"):
    return re.sub(r"\d{4}-\d\d-\d\d \d\d:\d\d:\d\d\.\d\d\d\d\d\d", replacement, text)


def test_isort_option_without_isort(git_repo, caplog):
    """Without isort, provide isort install instructions and error"""
    with isort_present(False), patch.object(
        darker.__main__, "isort", None
    ), pytest.raises(MissingPackageError) as exc_info:

        darker.__main__.main(["--isort", "."])

    assert (
        str(exc_info.value)
        == "Please run `pip install darker[isort]` to use the `--isort` option."
    )


@pytest.fixture
def run_isort(git_repo, monkeypatch, caplog, request, find_project_root_cache_clear):
    """Fixture for running Darker with requested arguments and a patched `isort`

    Provides an `run_isort.isort_code` mock object which allows checking whether and how
    the `isort.code()` function was called.

    """
    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({"test1.py": "original"}, commit="Initial commit")
    paths["test1.py"].write_bytes(b"changed")
    args = getattr(request, "param", ())
    isorted_code = "import os; import sys;"
    blacken_code = "import os\nimport sys\n"
    patch_run_black_ctx = patch.object(
        darker.__main__, "run_black", return_value=TextDocument(blacken_code)
    )
    with patch_run_black_ctx, patch("darker.import_sorting.isort_code") as isort_code:
        isort_code.return_value = isorted_code
        darker.__main__.main(["--isort", "./test1.py", *args])
        return SimpleNamespace(
            isort_code=darker.import_sorting.isort_code, caplog=caplog
        )


def test_isort_option_with_isort(run_isort):
    assert "Please run" not in run_isort.caplog.text


@pytest.mark.kwparametrize(
    dict(run_isort=(), isort_args={}),
    dict(run_isort=("--line-length", "120"), isort_args={"line_length": 120}),
    indirect=["run_isort"],
)
def test_isort_option_with_isort_calls_sortimports(tmpdir, run_isort, isort_args):
    run_isort.isort_code.assert_called_once_with(
        code="changed", settings_path=str(tmpdir), **isort_args
    )


A_PY = ["import sys", "import os", "print( '42')", ""]
A_PY_ISORT = ["import os", "import sys", "", "print( '42')", ""]
A_PY_BLACK = ["import sys", "import os", "", 'print("42")', ""]
A_PY_BLACK_UNNORMALIZE = ("import sys", "import os", "", "print('42')", "")
A_PY_BLACK_ISORT = ["import os", "import sys", "", 'print("42")', ""]

A_PY_DIFF_BLACK = [
    "--- a.py",
    "+++ a.py",
    "@@ -1,3 +1,4 @@",
    " import sys",
    " import os",
    "-print( '42')",
    "+",
    '+print("42")',
]

A_PY_DIFF_BLACK_NO_STR_NORMALIZE = [
    "--- a.py",
    "+++ a.py",
    "@@ -1,3 +1,4 @@",
    " import sys",
    " import os",
    "-print( '42')",
    "+",
    "+print('42')",
]

A_PY_DIFF_BLACK_ISORT = [
    "--- a.py",
    "+++ a.py",
    "@@ -1,3 +1,4 @@",
    "+import os",
    " import sys",
    "-import os",
    "-print( '42')",
    "+",
    '+print("42")',
]


@pytest.mark.kwparametrize(
    dict(
        isort_exclude={"**/*"},
        expect=[A_PY_BLACK],
    ),
    dict(
        expect=[A_PY_BLACK_ISORT],
    ),
    dict(
        black_config={"skip_string_normalization": True},
        isort_exclude={"**/*"},
        expect=[A_PY_BLACK_UNNORMALIZE],
    ),
    dict(
        black_exclude={Path("a.py")},
        isort_exclude={"**/*"},
        expect=[],
    ),
    dict(
        black_exclude={Path("a.py")},
        expect=[A_PY_ISORT],
    ),
    black_config={},
    black_exclude=set(),
    isort_exclude=set(),
)
@pytest.mark.parametrize("newline", ["\n", "\r\n"], ids=["unix", "windows"])
def test_format_edited_parts(
    git_repo, black_config, black_exclude, isort_exclude, newline, expect
):
    """Correct reformatting and import sorting changes are produced

    Black reformatting is done even if a file is excluded in Black configuration.
    File exclusion is done in Darker before calling
    :func:`~darker.__main__.format_edited_parts`.

    """
    paths = git_repo.add({"a.py": newline, "b.py": newline}, commit="Initial commit")
    paths["a.py"].write_bytes(newline.join(A_PY).encode("ascii"))
    paths["b.py"].write_bytes(f"print(42 ){newline}".encode("ascii"))

    result = darker.__main__.format_edited_parts(
        Path(git_repo.root),
        {Path("a.py")},
        Exclusions(black=black_exclude, isort=isort_exclude),
        RevisionRange("HEAD", ":WORKTREE:"),
        black_config,
        report_unmodified=False,
    )

    changes = [
        (path, worktree_content.string, chosen.string, chosen.lines)
        for path, worktree_content, chosen in result
    ]
    expect_changes = [
        (
            paths["a.py"],
            newline.join(A_PY),
            newline.join(expect_lines),
            tuple(expect_lines[:-1]),
        )
        for expect_lines in expect
    ]
    assert changes == expect_changes


def test_format_edited_parts_all_unchanged(git_repo, monkeypatch):
    """``format_edited_parts()`` yields nothing if no reformatting was needed"""
    monkeypatch.chdir(git_repo.root)
    paths = git_repo.add({"a.py": "pass\n", "b.py": "pass\n"}, commit="Initial commit")
    paths["a.py"].write_bytes(b'"properly"\n"formatted"\n')
    paths["b.py"].write_bytes(b'"not"\n"checked"\n')

    result = list(
        darker.__main__.format_edited_parts(
            Path(git_repo.root),
            {Path("a.py"), Path("b.py")},
            Exclusions(black=set(), isort=set()),
            RevisionRange("HEAD", ":WORKTREE:"),
            {},
            report_unmodified=False,
        )
    )

    assert result == []


def test_format_edited_parts_ast_changed(git_repo, caplog):
    """``darker.__main__.format_edited_parts()`` when reformatting changes the AST"""
    caplog.set_level(logging.DEBUG, logger="darker.__main__")
    paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
    paths["a.py"].write_bytes(b"8\n7\n6\n5\n4\n3\n2\n1\n")
    mock_ctx = patch.object(
        darker.verification.ASTVerifier, "is_equivalent_to_baseline", return_value=False
    )
    with mock_ctx, pytest.raises(NotEquivalentError):
        _ = list(
            darker.__main__.format_edited_parts(
                git_repo.root,
                {Path("a.py")},
                Exclusions(black=set(), isort={"**/*"}),
                RevisionRange("HEAD", ":WORKTREE:"),
                black_config={},
                report_unmodified=False,
            )
        )
    a_py = str(paths["a.py"])
    main = "darker.__main__:__main__.py"
    log = [
        line
        for line in re.sub(r":\d+", "", caplog.text).splitlines()
        if " lines of context " in line
    ]
    assert log == [
        f"DEBUG    {main} AST verification of {a_py} with 0 lines of context failed",
        f"DEBUG    {main} Trying with 5 lines of context for `git diff -U {a_py}`",
        f"DEBUG    {main} AST verification of {a_py} with 5 lines of context failed",
        f"DEBUG    {main} Trying with 7 lines of context for `git diff -U {a_py}`",
        f"DEBUG    {main} AST verification of {a_py} with 7 lines of context failed",
        f"DEBUG    {main} Trying with 8 lines of context for `git diff -U {a_py}`",
        f"DEBUG    {main} AST verification of {a_py} with 8 lines of context failed",
    ]


def test_format_edited_parts_isort_on_already_formatted(git_repo):
    """An already correctly formatted file after ``isort`` is simply skipped"""
    before = [
        "import a",
        "import b",
        "",
        "a.foo()",
        "b.bar()",
    ]
    after = [
        "import b",
        "",
        "b.bar()",
    ]
    paths = git_repo.add({"a.py": joinlines(before)}, commit="Initial commit")
    paths["a.py"].write_text(joinlines(after))

    result = darker.__main__.format_edited_parts(
        git_repo.root,
        {Path("a.py")},
        Exclusions(black=set(), isort=set()),
        RevisionRange("HEAD", ":WORKTREE:"),
        black_config={},
        report_unmodified=False,
    )

    assert list(result) == []


@pytest.mark.kwparametrize(
    dict(rev1="HEAD^", rev2="HEAD", expect=[]),
    dict(rev1="HEAD^", rev2=WORKTREE, expect=[(":WORKTREE:", "reformatted")]),
    dict(rev1="HEAD", rev2=WORKTREE, expect=[(":WORKTREE:", "reformatted")]),
)
def test_format_edited_parts_historical(git_repo, rev1, rev2, expect):
    """``format_edited_parts()`` is correct for different commit pairs"""
    a_py = {
        "HEAD^": TextDocument.from_lines(
            [
                "import a",
                "from b import bar, foo",
                "",
                "a.foo()",
                "bar()",
            ]
        ),
        "HEAD": TextDocument.from_lines(
            [
                "from b import bar, foo",
                "",
                "bar()",
            ]
        ),
        ":WORKTREE:": TextDocument.from_lines(
            [
                "from b import foo, bar",
                "",
                "bar( )",
            ]
        ),
        "reformatted": TextDocument.from_lines(
            [
                "from b import bar, foo",
                "",
                "bar()",
            ]
        ),
    }
    paths = git_repo.add({"a.py": a_py["HEAD^"].string}, commit="Initial commit")
    git_repo.add({"a.py": a_py["HEAD"].string}, commit="Modified a.py")
    paths["a.py"].write_text(a_py[":WORKTREE:"].string)

    result = darker.__main__.format_edited_parts(
        git_repo.root,
        {Path("a.py")},
        Exclusions(black=set(), isort=set()),
        RevisionRange(rev1, rev2),
        black_config={},
        report_unmodified=False,
    )

    assert list(result) == [(paths["a.py"], a_py[x[0]], a_py[x[1]]) for x in expect]


@pytest.mark.kwparametrize(
    dict(),
    dict(relative_path="file.py.12345.tmp"),
    dict(
        rev2_content="import  modified\n\nprint( original )\n",
        rev2_isorted="import  modified\n\nprint( original )\n",
        expect="import modified\n\nprint( original )\n",
    ),
    dict(
        rev2_content="import  original\n\nprint(modified )\n",
        rev2_isorted="import  original\n\nprint(modified )\n",
        expect="import  original\n\nprint(modified)\n",
    ),
    dict(
        relative_path="file.py.12345.tmp",
        rev2_content="import  modified\n\nprint( original )\n",
        rev2_isorted="import  modified\n\nprint( original )\n",
        expect="import modified\n\nprint( original )\n",
    ),
    dict(
        relative_path="file.py.12345.tmp",
        rev2_content="import  original\n\nprint(modified )\n",
        rev2_isorted="import  original\n\nprint(modified )\n",
        expect="import  original\n\nprint(modified)\n",
    ),
    relative_path="file.py",
    rev1="HEAD",
    rev2=":WORKTREE",
    rev2_content="import  original\nprint( original )\n",
    rev2_isorted="import  original\nprint( original )\n",
    enable_isort=False,
    black_config={},
    expect="import  original\nprint( original )\n",
)
def test_blacken_single_file(
    git_repo,
    relative_path,
    rev1,
    rev2,
    rev2_content,
    rev2_isorted,
    enable_isort,
    black_config,
    expect,
):
    """Test for ``_blacken_single_file``"""
    git_repo.add(
        {"file.py": "import  original\nprint( original )\n"}, commit="Initial commit"
    )
    result = darker.__main__._blacken_single_file(
        git_repo.root,
        Path(relative_path),
        Path("file.py"),
        EditedLinenumsDiffer(git_repo.root, RevisionRange(rev1, rev2)),
        TextDocument(rev2_content),
        TextDocument(rev2_isorted),
        enable_isort,
        black_config,
    )

    assert result.string == expect


@pytest.mark.kwparametrize(
    dict(arguments=["--diff"], expect_stdout=A_PY_DIFF_BLACK),
    dict(arguments=["--isort"], expect_a_py=A_PY_BLACK_ISORT),
    dict(
        arguments=["--skip-string-normalization", "--diff"],
        expect_stdout=A_PY_DIFF_BLACK_NO_STR_NORMALIZE,
    ),
    dict(arguments=[], expect_a_py=A_PY_BLACK, expect_retval=0),
    dict(arguments=["--isort", "--diff"], expect_stdout=A_PY_DIFF_BLACK_ISORT),
    dict(arguments=["--check"], expect_a_py=A_PY, expect_retval=1),
    dict(
        arguments=["--check", "--diff"],
        expect_stdout=A_PY_DIFF_BLACK,
        expect_retval=1,
    ),
    dict(arguments=["--check", "--isort"], expect_retval=1),
    dict(
        arguments=["--check", "--diff", "--isort"],
        expect_stdout=A_PY_DIFF_BLACK_ISORT,
        expect_retval=1,
    ),
    dict(
        arguments=["--check", "--lint", "echo subdir/a.py:1: message"],
        # Windows compatible path assertion using `pathlib.Path()`
        expect_stdout=["", f"subdir/a.py:1: message {Path('/subdir/a.py')}"],
        expect_retval=1,
    ),
    dict(
        arguments=["--diff", "--lint", "echo subdir/a.py:1: message"],
        # Windows compatible path assertion using `pathlib.Path()`
        expect_stdout=A_PY_DIFF_BLACK
        + ["", f"subdir/a.py:1: message {Path('/subdir/a.py')}"],
        expect_retval=1,
    ),
    dict(
        arguments=[],
        pyproject_toml="""
           [tool.black]
           exclude = 'a.py'
           """,
        expect_a_py=A_PY,
    ),
    dict(
        arguments=["--diff"],
        pyproject_toml="""
           [tool.black]
           exclude = 'a.py'
           """,
        expect_stdout=[],
    ),
    dict(
        arguments=[],
        pyproject_toml="""
           [tool.black]
           extend_exclude = 'a.py'
           """,
        expect_a_py=A_PY,
    ),
    dict(
        arguments=["--diff"],
        pyproject_toml="""
           [tool.black]
           extend_exclude = 'a.py'
           """,
        expect_stdout=[],
    ),
    dict(
        arguments=[],
        pyproject_toml="""
           [tool.black]
           force_exclude = 'a.py'
           """,
        expect_a_py=A_PY,
    ),
    dict(
        arguments=["--diff"],
        pyproject_toml="""
           [tool.black]
           force_exclude = 'a.py'
           """,
        expect_stdout=[],
    ),
    dict(
        arguments=["--diff"],
        expect_stdout=A_PY_DIFF_BLACK,
        root_as_cwd=False,
    ),
    # for all test cases, by default there's no output, `a.py` stays unmodified, and the
    # return value is a zero:
    pyproject_toml="",
    expect_stdout=[],
    expect_a_py=A_PY,
    expect_retval=0,
    root_as_cwd=True,
)
@pytest.mark.parametrize("newline", ["\n", "\r\n"], ids=["unix", "windows"])
def test_main(
    git_repo,
    monkeypatch,
    capsys,
    find_project_root_cache_clear,
    arguments,
    newline,
    pyproject_toml,
    expect_stdout,
    expect_a_py,
    expect_retval,
    root_as_cwd,
    tmp_path_factory,
):
    """Main function outputs diffs and modifies files correctly"""
    if root_as_cwd:
        cwd = git_repo.root
        pwd = Path("")
    else:
        cwd = tmp_path_factory.mktemp("not_a_git_repo")
        pwd = git_repo.root
    monkeypatch.chdir(cwd)
    paths = git_repo.add(
        {
            "pyproject.toml": dedent(pyproject_toml),
            "subdir/a.py": newline,
            "b.py": newline,
        },
        commit="Initial commit",
    )
    paths["subdir/a.py"].write_bytes(newline.join(A_PY).encode("ascii"))
    paths["b.py"].write_bytes(f"print(42 ){newline}".encode("ascii"))

    retval = darker.__main__.main(arguments + [str(pwd / "subdir")])

    stdout = capsys.readouterr().out.replace(str(git_repo.root), "")
    diff_output = stdout.splitlines(False)
    if expect_stdout:
        if "--diff" in arguments:
            assert "\t" in diff_output[0], diff_output[0]
            diff_output[0], old_mtime = diff_output[0].split("\t", 1)
            assert old_mtime.endswith(" +0000")
            assert "\t" in diff_output[1], diff_output[1]
            diff_output[1], new_mtime = diff_output[1].split("\t", 1)
            assert new_mtime.endswith(" +0000")
            assert all("\t" not in line for line in diff_output[2:])
        else:
            assert all("\t" not in line for line in diff_output)
    assert diff_output == expect_stdout
    assert paths["subdir/a.py"].read_bytes().decode("ascii") == newline.join(
        expect_a_py
    )
    assert paths["b.py"].read_bytes().decode("ascii") == f"print(42 ){newline}"
    assert retval == expect_retval


def test_main_in_plain_directory(tmp_path, capsys):
    """Darker works also in a plain directory tree"""
    subdir_a = tmp_path / "subdir_a"
    subdir_c = tmp_path / "subdir_b/subdir_c"
    subdir_a.mkdir()
    subdir_c.mkdir(parents=True)
    (subdir_a / "non-python file.txt").write_text("not  reformatted\n")
    (subdir_a / "python file.py").write_text("import  sys, os\nprint('ok')")
    (subdir_c / "another python file.py").write_text("a  =5")
    with patch.object(darker.linting, "run_linter") as run_linter:

        retval = darker.__main__.main(
            ["--diff", "--check", "--isort", "--lint", "my-linter", str(tmp_path)]
        )

    assert retval == 1
    assert run_linter.call_args_list == [
        call(
            "my-linter",
            tmp_path,
            {
                Path("subdir_a/python file.py"),
                Path("subdir_b/subdir_c/another python file.py"),
            },
            RevisionRange(rev1="HEAD", rev2=":WORKTREE:"),
            False,
        )
    ]
    output = capsys.readouterr().out
    output = _replace_diff_timestamps(output)
    assert output == dedent(
        """\
        --- subdir_a/python file.py	<timestamp> +0000
        +++ subdir_a/python file.py	<timestamp> +0000
        @@ -1,2 +1,4 @@
        -import  sys, os
        -print('ok')
        +import os
        +import sys
        +
        +print("ok")
        --- subdir_b/subdir_c/another python file.py	<timestamp> +0000
        +++ subdir_b/subdir_c/another python file.py	<timestamp> +0000
        @@ -1 +1 @@
        -a  =5
        +a = 5
        """
    )


@pytest.mark.parametrize(
    "encoding, text", [(b"utf-8", b"touch\xc3\xa9"), (b"iso-8859-1", b"touch\xe9")]
)
@pytest.mark.parametrize("newline", [b"\n", b"\r\n"])
def test_main_encoding(
    git_repo, find_project_root_cache_clear, encoding, text, newline
):
    """Encoding and newline of the file is kept unchanged after reformatting"""
    paths = git_repo.add({"a.py": newline.decode("ascii")}, commit="Initial commit")
    edited = [b"# coding: ", encoding, newline, b's="', text, b'"', newline]
    expect = [b"# coding: ", encoding, newline, b's = "', text, b'"', newline]
    paths["a.py"].write_bytes(b"".join(edited))

    retval = darker.__main__.main(["a.py"])

    result = paths["a.py"].read_bytes()
    assert retval == 0
    assert result == b"".join(expect)


def test_main_historical(git_repo):
    """Stop if rev2 isn't the working tree and no ``--diff`` or ``--check`` provided"""
    with pytest.raises(ArgumentError):

        darker.__main__.main(["--revision=foo..bar", "."])


@pytest.mark.parametrize("arguments", [["--diff"], ["--check"], ["--diff", "--check"]])
@pytest.mark.parametrize("src", [".", "foo/..", "{git_repo_root}"])
def test_main_historical_ok(git_repo, arguments, src):
    """Runs ok for repository root with rev2 specified and ``--diff`` or ``--check``"""
    git_repo.add({"README": "first"}, commit="Initial commit")
    initial = git_repo.get_hash()
    git_repo.add({"README": "second"}, commit="Second commit")
    second = git_repo.get_hash()

    darker.__main__.main(
        arguments
        + [f"--revision={initial}..{second}", src.format(git_repo_root=git_repo.root)]
    )


def test_main_pre_commit_head(git_repo, monkeypatch):
    """Warn if run by pre-commit, rev2=HEAD and no ``--diff`` or ``--check`` provided"""
    git_repo.add({"a.py": "original = 1"}, commit="Add a.py")
    initial = git_repo.get_hash()
    git_repo.add({"a.py": "modified  = 2"}, commit="Modify a.py")
    monkeypatch.setenv("PRE_COMMIT_FROM_REF", initial)
    monkeypatch.setenv("PRE_COMMIT_TO_REF", "HEAD")
    with pytest.warns(
        UserWarning,
        match=re.escape(
            "Darker was called by pre-commit, comparing HEAD to an older commit."
            " As an experimental feature, allowing overwriting of files."
            " See https://github.com/akaihola/darker/issues/180 for details."
        ),
    ):

        result = darker.__main__.main(["--revision=:PRE-COMMIT:", "a.py"])

    assert result == 0


def test_main_historical_pre_commit(git_repo, monkeypatch):
    """Stop if run by pre-commit, rev2 older than HEAD and no ``--diff``/``--check``"""
    git_repo.add({"README.txt": ""}, commit="Initial commit")
    initial = git_repo.get_hash()
    git_repo.add({"a.py": "original"}, commit="Add a.py")
    older_commit = git_repo.get_hash()
    git_repo.add({"a.py": "modified"}, commit="Modify a.py")
    monkeypatch.setenv("PRE_COMMIT_FROM_REF", initial)
    monkeypatch.setenv("PRE_COMMIT_TO_REF", older_commit)
    with pytest.raises(
        ArgumentError,
        match=(
            re.escape(
                f"Can't write reformatted files for revision '{older_commit}'."
                " Either --diff or --check must be used."
            )
        ),
    ):

        darker.__main__.main(["--revision=:PRE-COMMIT:", "a.py"])


def test_main_vscode_tmpfile(git_repo, capsys):
    """Main function handles VSCode `.py.<HASH>.tmp` files correctly"""
    _ = git_repo.add(
        {"a.py": "print ( 'reformat me' ) \n"},
        commit="Initial commit",
    )
    (git_repo.root / "a.py.hash.tmp").write_text("print ( 'reformat me now' ) \n")

    retval = darker.__main__.main(["--diff", "a.py.hash.tmp"])

    assert retval == 0
    outerr = capsys.readouterr()
    assert outerr.err == ""
    stdout = _replace_diff_timestamps(outerr.out.replace(str(git_repo.root), ""))
    diff_output = stdout.splitlines(False)
    assert diff_output == [
        "--- a.py.hash.tmp\t<timestamp> +0000",
        "+++ a.py.hash.tmp\t<timestamp> +0000",
        "@@ -1 +1 @@",
        "-print ( 'reformat me now' ) ",
        '+print("reformat me now")',
    ]


def test_print_diff(tmp_path, capsys):
    """print_diff() prints Black-style diff output with 5 lines of context"""
    Path(tmp_path / "a.py").write_text("dummy\n", encoding="utf-8")
    darker.__main__.print_diff(
        Path(tmp_path / "a.py"),
        TextDocument.from_lines(
            [
                "unchanged",
                "removed",
                "kept 1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "10",
                "11",
                "changed",
            ],
            mtime="2020-10-08 19:16:22.146405 +0000",
        ),
        TextDocument.from_lines(
            [
                "inserted",
                "unchanged",
                "kept 1",
                "2",
                "3",
                "4",
                "5",
                "6",
                "7",
                "8",
                "9",
                "10",
                "11",
                "Changed",
            ],
            mtime="2020-10-08 19:21:09.005501 +0000",
        ),
        root=tmp_path,
        use_color=False,
    )

    assert capsys.readouterr().out.splitlines() == [
        "--- a.py\t2020-10-08 19:16:22.146405 +0000",
        "+++ a.py\t2020-10-08 19:21:09.005501 +0000",
        "@@ -1,7 +1,7 @@",
        "+inserted",
        " unchanged",
        "-removed",
        " kept 1",
        " 2",
        " 3",
        " 4",
        " 5",
        "@@ -9,6 +9,6 @@",
        " 7",
        " 8",
        " 9",
        " 10",
        " 11",
        "-changed",
        "+Changed",
    ]


@pytest.mark.kwparametrize(
    dict(new_content=TextDocument(), expect=b""),
    dict(new_content=TextDocument(lines=["touché"]), expect=b"touch\xc3\xa9\n"),
    dict(
        new_content=TextDocument(lines=["touché"], newline="\r\n"),
        expect=b"touch\xc3\xa9\r\n",
    ),
    dict(
        new_content=TextDocument(lines=["touché"], encoding="iso-8859-1"),
        expect=b"touch\xe9\n",
    ),
)
def test_modify_file(tmp_path, new_content, expect):
    """Encoding and newline are respected when writing a text file on disk"""
    path = tmp_path / "test.py"

    darker.__main__.modify_file(path, new_content)

    result = path.read_bytes()
    assert result == expect


@pytest.mark.kwparametrize(
    dict(
        new_content=TextDocument(lines=['print("foo")']),
        use_color=False,
        expect=('print("foo")\n',),
    ),
    dict(
        new_content=TextDocument(lines=['print("foo")']),
        use_color=False,
        expect=('print("foo")\n',),
    ),
    dict(
        new_content=TextDocument(lines=['print("foo")']),
        use_color=True,
        expect=(
            '\x1b[36mprint\x1b[39;49;00m(\x1b[33m"\x1b[39;49;00mfoo'
            + '\x1b[33m"\x1b[39;49;00m)\n',
            '\x1b[36mprint\x1b[39;49;00m(\x1b[33m"\x1b[39;49;00m\x1b[33mfoo'
            + '\x1b[39;49;00m\x1b[33m"\x1b[39;49;00m)\n',
            # Pygments 2.4.x:
            '\x1b[34mprint\x1b[39;49;00m(\x1b[33m"\x1b[39;49;00mfoo'
            + '\x1b[33m"\x1b[39;49;00m)\n',
        ),
    ),
)
def test_print_source(new_content, use_color, expect, capsys):
    """Highlight is applied only if specified, final newline is handled correctly."""
    darker.__main__.print_source(new_content, use_color=use_color)

    assert capsys.readouterr().out in expect


def test_stdout_path_resolution(git_repo, capsys):
    """When using ``--stdout``, file paths are resolved correctly"""
    git_repo.add({"src/menu.py": "print ( 'foo' )\n"})

    result = darker.__main__.main(["--stdout", "./src/menu.py"])

    assert result == 0
    assert capsys.readouterr().out == 'print("foo")\n'



import re
from typing import Pattern

from setuptools import setup

SIDEBAR_RE = re.compile(r"\+---.*:alt: Support", re.DOTALL)
CONTRIBUTORS_RE = re.compile(r"""\nThanks goes .*\nThis project follows """, re.DOTALL)


def replace(name: str, regex: Pattern[str], replacement: str, content: str) -> str:
    """Replace/remove a section from the package description, based on a regex

    Raise an exception if the regular expression doesn't match anything.

    """
    modified_content = regex.sub(replacement, content)
    if modified_content != content:
        return modified_content
    raise RuntimeError(
        f"The {name} wasn't found in README.rst using the pattern '{regex.pattern}'"
    )


def make_pypi_compliant_readme() -> str:
    """Remove raw HTML from ``README.rst`` before packaging

    The sidebar and the contributors table should only be shown on GitHub. They must be
    removed before being displayed on PyPI. The simplest way to do this is to simply
    match and strip it when creating a distribution archive.

    This function reads the contents of ``README.rst``, removes the sidebar, replaces
    the contributors raw HTML section with a plain-text link to the README on GitHub and
    returns the resulting string.

    :return: The contents of a PyPI compliant ``README.rst``

    """
    with open("README.rst", encoding="utf-8") as fp:
        original_readme = fp.read()
    no_sidebar_readme = replace("sidebar", SIDEBAR_RE, "", original_readme)
    return replace(
        "contributors table",
        CONTRIBUTORS_RE,
        "\nSee README.rst_ for the list of contributors.\n\nThis project follows ",
        no_sidebar_readme,
    )


setup(long_description=make_pypi_compliant_readme())



"""Tests for `darker.config`"""

# pylint: disable=unused-argument

import os
import re
from argparse import ArgumentParser, Namespace
from pathlib import Path
from textwrap import dedent

import pytest

from darker.config import (
    ConfigurationError,
    DarkerConfig,
    OutputMode,
    TomlArrayLinesEncoder,
    dump_config,
    get_effective_config,
    get_modified_config,
    load_config,
    replace_log_level_name,
)
from darker.tests.helpers import raises_if_exception


@pytest.mark.kwparametrize(
    dict(list_value=[], expect="[\n]"),
    dict(list_value=["one value"], expect='[\n    "one value",\n]'),
    dict(list_value=["two", "values"], expect='[\n    "two",\n    "values",\n]'),
    dict(
        list_value=[
            "a",
            "dozen",
            "short",
            "string",
            "values",
            "in",
            "the",
            "list",
            "of",
            "strings",
            "to",
            "format",
        ],
        expect=(
            '[\n    "a",\n    "dozen",\n    "short",\n    "string",\n    "values"'
            ',\n    "in",\n    "the",\n    "list",\n    "of",\n    "strings"'
            ',\n    "to",\n    "format",\n]'
        ),
    ),
)
def test_toml_array_lines_encoder(list_value, expect):
    result = TomlArrayLinesEncoder().dump_list(list_value)

    assert result == expect


@pytest.mark.kwparametrize(
    dict(log_level=0, expect="NOTSET"),
    dict(log_level=10, expect="DEBUG"),
    dict(log_level=20, expect="INFO"),
    dict(log_level=30, expect="WARNING"),
    dict(log_level=40, expect="ERROR"),
    dict(log_level=50, expect="CRITICAL"),
    dict(log_level="DEBUG", expect=10),
    dict(log_level="INFO", expect=20),
    dict(log_level="WARNING", expect=30),
    dict(log_level="WARN", expect=30),
    dict(log_level="ERROR", expect=40),
    dict(log_level="CRITICAL", expect=50),
    dict(log_level="FOOBAR", expect="Level FOOBAR"),
)
def test_replace_log_level_name(log_level, expect):
    config = DarkerConfig() if log_level is None else DarkerConfig(log_level=log_level)
    replace_log_level_name(config)

    assert config["log_level"] == expect


@pytest.mark.kwparametrize(
    dict(diff=False, stdout=False, expect=None),
    dict(diff=False, stdout=True, expect=None),
    dict(diff=True, stdout=False, expect=None),
    dict(diff=True, stdout=True, expect=ConfigurationError),
)
def test_output_mode_validate_diff_stdout(diff, stdout, expect):
    """Validation fails only if ``--diff`` and ``--stdout`` are both enabled"""
    with raises_if_exception(expect):
        OutputMode.validate_diff_stdout(diff, stdout)


@pytest.mark.kwparametrize(
    dict(stdout=False, src=[], expect=None),
    dict(stdout=False, src=["first.py"], expect=None),
    dict(stdout=False, src=["first.py", "second.py"], expect=None),
    dict(stdout=False, src=["first.py", "missing.py"], expect=None),
    dict(stdout=False, src=["missing.py"], expect=None),
    dict(stdout=False, src=["missing.py", "another_missing.py"], expect=None),
    dict(stdout=False, src=["directory"], expect=None),
    dict(stdout=True, src=[], expect=ConfigurationError),
    dict(stdout=True, src=["first.py"], expect=None),
    dict(stdout=True, src=["first.py", "second.py"], expect=ConfigurationError),
    dict(stdout=True, src=["first.py", "missing.py"], expect=ConfigurationError),
    dict(stdout=True, src=["missing.py"], expect=ConfigurationError),
    dict(stdout=True, src=["missing.py", "another.py"], expect=ConfigurationError),
    dict(stdout=True, src=["directory"], expect=ConfigurationError),
)
def test_output_mode_validate_stdout_src(tmp_path, monkeypatch, stdout, expect, src):
    """Validation fails only if exactly one file isn't provided for ``--stdout``"""
    monkeypatch.chdir(tmp_path)
    Path("first.py").touch()
    Path("second.py").touch()
    with raises_if_exception(expect):

        OutputMode.validate_stdout_src(stdout, src)


@pytest.mark.kwparametrize(
    dict(diff=False, stdout=False, expect="NOTHING"),
    dict(diff=False, stdout=True, expect="CONTENT"),
    dict(diff=True, stdout=False, expect="DIFF"),
    dict(diff=True, stdout=True, expect=ConfigurationError),
)
def test_output_mode_from_args(diff, stdout, expect):
    """Correct output mode results from the ``--diff`` and ``stdout`` options"""
    args = Namespace()
    args.diff = diff
    args.stdout = stdout
    with raises_if_exception(expect):

        result = OutputMode.from_args(args)

        assert result == expect


@pytest.mark.kwparametrize(
    dict(),
    dict(cwd="lvl1"),
    dict(cwd="lvl1/lvl2"),
    dict(cwd="has_git", expect={}),
    dict(cwd="has_git/lvl1", expect={}),
    dict(cwd="has_pyp", expect={"CONFIG_PATH": "has_pyp"}),
    dict(cwd="has_pyp/lvl1", expect={"CONFIG_PATH": "has_pyp"}),
    dict(srcs=["root.py"]),
    dict(srcs=["../root.py"], cwd="lvl1"),
    dict(srcs=["../root.py"], cwd="has_git"),
    dict(srcs=["../root.py"], cwd="has_pyp"),
    dict(srcs=["root.py", "lvl1/lvl1.py"]),
    dict(srcs=["../root.py", "lvl1.py"], cwd="lvl1"),
    dict(srcs=["../root.py", "../lvl1/lvl1.py"], cwd="has_git"),
    dict(srcs=["../root.py", "../lvl1/lvl1.py"], cwd="has_pyp"),
    dict(srcs=["has_pyp/pyp.py", "lvl1/lvl1.py"]),
    dict(srcs=["../has_pyp/pyp.py", "lvl1.py"], cwd="lvl1"),
    dict(srcs=["../has_pyp/pyp.py", "../lvl1/lvl1.py"], cwd="has_git"),
    dict(srcs=["pyp.py", "../lvl1/lvl1.py"], cwd="has_pyp"),
    dict(
        srcs=["has_pyp/lvl1/l1.py", "has_pyp/lvl1b/l1b.py"],
        expect={"CONFIG_PATH": "has_pyp"},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="lvl1",
        expect={"CONFIG_PATH": "has_pyp"},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="has_git",
        expect={"CONFIG_PATH": "has_pyp"},
    ),
    dict(
        srcs=["lvl1/l1.py", "lvl1b/l1b.py"],
        cwd="has_pyp",
        expect={"CONFIG_PATH": "has_pyp"},
    ),
    dict(
        srcs=["full_example/full.py"],
        expect={
            "check": True,
            "diff": True,
            "isort": True,
            "lint": ["flake8", "mypy", "pylint"],
            "log_level": 10,
            "revision": "main",
            "src": ["src", "tests"],
        },
    ),
    dict(srcs=["stdout_example/dummy.py"], expect={"stdout": True}),
    dict(confpath="c", expect={"PYP_TOML": 1}),
    dict(confpath="c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="lvl1", confpath="../c", expect={"PYP_TOML": 1}),
    dict(cwd="lvl1", confpath="../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="lvl1/lvl2", confpath="../../c", expect={"PYP_TOML": 1}),
    dict(cwd="lvl1/lvl2", confpath="../../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="has_git", confpath="../c", expect={"PYP_TOML": 1}),
    dict(cwd="has_git", confpath="../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="has_git/lvl1", confpath="../../c", expect={"PYP_TOML": 1}),
    dict(cwd="has_git/lvl1", confpath="../../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="has_pyp", confpath="../c", expect={"PYP_TOML": 1}),
    dict(cwd="has_pyp", confpath="../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(cwd="has_pyp/lvl1", confpath="../../c", expect={"PYP_TOML": 1}),
    dict(cwd="has_pyp/lvl1", confpath="../../c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(srcs=["root.py"], confpath="c", expect={"PYP_TOML": 1}),
    dict(srcs=["root.py"], confpath="c/pyproject.toml", expect={"PYP_TOML": 1}),
    dict(srcs=["../root.py"], cwd="lvl1", confpath="../c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["../root.py"],
        cwd="lvl1",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["../root.py"], cwd="has_git", confpath="../c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["../root.py"],
        cwd="has_git",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["../root.py"], cwd="has_pyp", confpath="../c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["../root.py"],
        cwd="has_pyp",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["root.py", "lvl1/lvl1.py"], confpath="c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["root.py", "lvl1/lvl1.py"],
        confpath="c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "lvl1.py"],
        cwd="lvl1",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "lvl1.py"],
        cwd="lvl1",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "../lvl1/lvl1.py"],
        cwd="has_git",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "../lvl1/lvl1.py"],
        cwd="has_git",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "../lvl1/lvl1.py"],
        cwd="has_pyp",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../root.py", "../lvl1/lvl1.py"],
        cwd="has_pyp",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["has_pyp/pyp.py", "lvl1/lvl1.py"], confpath="c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["has_pyp/pyp.py", "lvl1/lvl1.py"],
        confpath="c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/pyp.py", "lvl1.py"],
        cwd="lvl1",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/pyp.py", "lvl1.py"],
        cwd="lvl1",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/pyp.py", "../lvl1/lvl1.py"],
        cwd="has_git",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/pyp.py", "../lvl1/lvl1.py"],
        cwd="has_git",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["pyp.py", "../lvl1/lvl1.py"],
        cwd="has_pyp",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["pyp.py", "../lvl1/lvl1.py"],
        cwd="has_pyp",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["has_pyp/lvl1/l1.py", "has_pyp/lvl1b/l1b.py"],
        confpath="c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["has_pyp/lvl1/l1.py", "has_pyp/lvl1b/l1b.py"],
        confpath="c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="lvl1",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="lvl1",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="has_git",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["../has_pyp/lvl1/l1.py", "../has_pyp/lvl1b/l1b.py"],
        cwd="has_git",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["lvl1/l1.py", "lvl1b/l1b.py"],
        cwd="has_pyp",
        confpath="../c",
        expect={"PYP_TOML": 1},
    ),
    dict(
        srcs=["lvl1/l1.py", "lvl1b/l1b.py"],
        cwd="has_pyp",
        confpath="../c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["full_example/full.py"], confpath="c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["full_example/full.py"],
        confpath="c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    dict(srcs=["stdout_example/dummy.py"], confpath="c", expect={"PYP_TOML": 1}),
    dict(
        srcs=["stdout_example/dummy.py"],
        confpath="c/pyproject.toml",
        expect={"PYP_TOML": 1},
    ),
    srcs=[],
    cwd=".",
    confpath=None,
    expect={"CONFIG_PATH": "."},
)
def test_load_config(
    find_project_root_cache_clear, tmp_path, monkeypatch, srcs, cwd, confpath, expect
):
    (tmp_path / ".git").mkdir()
    (tmp_path / "pyproject.toml").write_text('[tool.darker]\nCONFIG_PATH = "."\n')
    (tmp_path / "lvl1/lvl2").mkdir(parents=True)
    (tmp_path / "has_git/.git").mkdir(parents=True)
    (tmp_path / "has_git/lvl1").mkdir()
    (tmp_path / "has_pyp/lvl1").mkdir(parents=True)
    (tmp_path / "has_pyp/pyproject.toml").write_text(
        '[tool.darker]\nCONFIG_PATH = "has_pyp"\n'
    )
    (tmp_path / "full_example").mkdir()
    (tmp_path / "full_example/pyproject.toml").write_text(
        dedent(
            """
            [tool.darker]
            src = [
                "src",
                "tests",
            ]
            revision = "main"
            diff = true
            check = true
            isort = true
            lint = [
                "flake8",
                "mypy",
                "pylint",
            ]
            log_level = "DEBUG"
            """
        )
    )
    (tmp_path / "stdout_example").mkdir()
    (tmp_path / "stdout_example/pyproject.toml").write_text(
        "[tool.darker]\nstdout = true\n"
    )
    (tmp_path / "c").mkdir()
    (tmp_path / "c" / "pyproject.toml").write_text("[tool.darker]\nPYP_TOML = 1\n")
    monkeypatch.chdir(tmp_path / cwd)

    result = load_config(confpath, srcs)

    assert result == expect


@pytest.mark.kwparametrize(
    dict(path=".", expect="Configuration file pyproject.toml not found"),
    dict(path="./foo.toml", expect="Configuration file ./foo.toml not found"),
    dict(
        path="empty", expect=f"Configuration file empty{os.sep}pyproject.toml not found"
    ),
    dict(
        path="empty/",
        expect=f"Configuration file empty{os.sep}pyproject.toml not found",
    ),
    dict(path="subdir/foo.toml", expect="Configuration file subdir/foo.toml not found"),
    dict(
        path="missing_dir",
        expect="Configuration file missing_dir not found",
    ),
    dict(
        path=f"missing_dir{os.sep}",
        expect=f"Configuration file missing_dir{os.sep}pyproject.toml not found",
    ),
    dict(
        path="missing_dir/foo.toml",
        expect="Configuration file missing_dir/foo.toml not found",
    ),
)
def test_load_config_explicit_path_errors(tmp_path, monkeypatch, path, expect):
    """``load_config()`` raises an error if given path is not a file"""
    monkeypatch.chdir(tmp_path)
    (tmp_path / "subdir").mkdir()
    (tmp_path / "subdir" / "pyproject.toml").write_text("")
    (tmp_path / "empty").mkdir()
    with pytest.raises(ConfigurationError, match=re.escape(expect)):

        _ = load_config(path, ["."])


@pytest.mark.kwparametrize(
    dict(args=Namespace(), expect={}),
    dict(args=Namespace(one="option"), expect={"one": "option"}),
    dict(args=Namespace(log_level=10), expect={"log_level": "DEBUG"}),
    dict(
        args=Namespace(two="options", log_level=20),
        expect={"two": "options", "log_level": "INFO"},
    ),
    dict(args=Namespace(diff=True, stdout=True), expect=ConfigurationError),
)
def test_get_effective_config(args, expect):
    """``get_effective_config()`` converts command line options correctly"""
    with raises_if_exception(expect):

        result = get_effective_config(args)

        assert result == expect


@pytest.mark.kwparametrize(
    dict(args=Namespace(), expect={}),
    dict(args=Namespace(unknown="option"), expect={"unknown": "option"}),
    dict(args=Namespace(log_level=10), expect={"log_level": "DEBUG"}),
    dict(args=Namespace(names=[], int=42, string="fourty-two"), expect={"names": []}),
    dict(
        args=Namespace(names=["bar"], int=42, string="fourty-two"),
        expect={"names": ["bar"]},
    ),
    dict(
        args=Namespace(names=["foo"], int=43, string="fourty-two"), expect={"int": 43}
    ),
    dict(args=Namespace(names=["foo"], int=42, string="one"), expect={"string": "one"}),
)
def test_get_modified_config(args, expect):
    parser = ArgumentParser()
    parser.add_argument("names", nargs="*", default=["foo"])
    parser.add_argument("--int", dest="int", default=42)
    parser.add_argument("--string", default="fourty-two")
    result = get_modified_config(parser, args)

    assert result == expect


@pytest.mark.kwparametrize(
    dict(config={}, expect="[tool.darker]\n"),
    dict(config={"str": "value"}, expect='[tool.darker]\nstr = "value"\n'),
    dict(config={"int": 42}, expect="[tool.darker]\nint = 42\n"),
    dict(config={"float": 4.2}, expect="[tool.darker]\nfloat = 4.2\n"),
    dict(
        config={"list": ["foo", "bar"]},
        expect=dedent(
            """\
            [tool.darker]
            list = [
                "foo",
                "bar",
            ]
            """
        ),
    ),
    dict(
        config={
            "src": ["main.py"],
            "revision": "master",
            "diff": False,
            "stdout": False,
            "check": False,
            "isort": False,
            "lint": [],
            "config": None,
            "log_level": "DEBUG",
            "skip_string_normalization": None,
            "line_length": None,
        },
        expect=dedent(
            """\
            [tool.darker]
            src = [
                "main.py",
            ]
            revision = "master"
            diff = false
            stdout = false
            check = false
            isort = false
            lint = [
            ]
            log_level = "DEBUG"
            """
        ),
    ),
)
def test_dump_config(config, expect):
    """``dump_config()`` outputs configuration correctly in the TOML format"""
    result = dump_config(config)

    assert result == expect



"""Load and save configuration in TOML format"""

import logging
import os
import sys
from argparse import ArgumentParser, Namespace
from dataclasses import dataclass
from pathlib import Path
from typing import Collection, Iterable, List, Optional, cast

import toml

from darker.black_compat import find_project_root

if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    from typing_extensions import TypedDict


class TomlArrayLinesEncoder(toml.TomlEncoder):  # type: ignore
    """Format TOML so list items are each on their own line"""

    def dump_list(self, v: Iterable[object]) -> str:
        """Format a list value"""
        return "[{}\n]".format("".join(f"\n    {self.dump_value(item)}," for item in v))


class DarkerConfig(TypedDict, total=False):
    """Dictionary representing ``[tool.darker]`` from ``pyproject.toml``"""

    src: List[str]
    revision: str
    diff: bool
    stdout: bool
    check: bool
    isort: bool
    lint: List[str]
    config: str
    log_level: int
    color: bool
    skip_string_normalization: bool
    skip_magic_trailing_comma: bool
    line_length: int
    workers: int


class OutputMode:
    """The output mode to use: all file content, just the diff, or no output"""

    NOTHING = "NOTHING"
    DIFF = "DIFF"
    CONTENT = "CONTENT"

    @classmethod
    def from_args(cls, args: Namespace) -> str:
        """Resolve output mode based on  ``diff`` and ``stdout`` options"""
        OutputMode.validate_diff_stdout(args.diff, args.stdout)
        if args.diff:
            return cls.DIFF
        if args.stdout:
            return cls.CONTENT
        return cls.NOTHING

    @staticmethod
    def validate_diff_stdout(diff: bool, stdout: bool) -> None:
        """Raise an exception if ``diff`` and ``stdout`` options are both enabled"""
        if diff and stdout:
            raise ConfigurationError(
                "The `diff` and `stdout` options can't both be enabled"
            )

    @staticmethod
    def validate_stdout_src(stdout: bool, src: List[str]) -> None:
        """Raise an exception in ``stdout`` mode if not exactly one path is provided"""
        if not stdout:
            return
        if len(src) == 1 and Path(src[0]).is_file():
            return
        raise ConfigurationError(
            "Exactly one Python source file which exists on disk must be provided when"
            " using the `stdout` option"
        )


class ConfigurationError(Exception):
    """Exception class for invalid configuration values"""


def replace_log_level_name(config: DarkerConfig) -> None:
    """Replace numeric log level in configuration with the name of the log level"""
    if "log_level" in config:
        config["log_level"] = logging.getLevelName(config["log_level"])


def validate_config_output_mode(config: DarkerConfig) -> None:
    """Make sure both ``diff`` and ``stdout`` aren't enabled in configuration"""
    OutputMode.validate_diff_stdout(
        config.get("diff", False), config.get("stdout", False)
    )


def override_color_with_environment(pyproject_config: DarkerConfig) -> DarkerConfig:
    """Override ``color`` if the ``PY_COLORS`` environment variable is '0' or '1'

    :param config: The configuration read from ``pyproject.toml``
    :return: The modified configuration

    """
    config = pyproject_config.copy()
    py_colors = os.getenv("PY_COLORS")
    if py_colors in {"0", "1"}:
        config["color"] = py_colors == "1"
    elif os.getenv("NO_COLOR") is not None:
        config["color"] = False
    elif os.getenv("FORCE_COLOR") is not None:
        config["color"] = True
    return config


def load_config(path: Optional[str], srcs: Iterable[str]) -> DarkerConfig:
    """Find and load Darker configuration from a TOML configuration file

    Darker determines the location for the configuration file by trying the following:
    - the file path in the `path` argument, given using the ``-c``/``--config`` command
      line option
    - ``pyproject.toml`` inside the directory specified by the `path` argument
    - ``pyproject.toml`` from a common parent directory to all items in `srcs`
    - ``pyproject.toml`` in the current working directory if `srcs` is empty

    :param path: The file or directory specified using the ``-c``/``--config`` command
                 line option, or `None` if the option was omitted.
    :param srcs: File(s) and directory/directories to be processed by Darker.

    """
    if path:
        for candidate_path in [Path(path), Path(path, "pyproject.toml")]:
            if candidate_path.is_file():
                config_path = candidate_path
                break
        else:
            if Path(path).is_dir() or path.endswith(os.sep):
                raise ConfigurationError(
                    f"Configuration file {Path(path, 'pyproject.toml')} not found"
                )
            raise ConfigurationError(f"Configuration file {path} not found")
    else:
        config_path = find_project_root(tuple(srcs or ["."])) / "pyproject.toml"
        if not config_path.is_file():
            return {}
    pyproject_toml = toml.load(config_path)
    config = cast(DarkerConfig, pyproject_toml.get("tool", {}).get("darker", {}) or {})
    replace_log_level_name(config)
    validate_config_output_mode(config)
    return config


def get_effective_config(args: Namespace) -> DarkerConfig:
    """Return all configuration options"""
    config = cast(DarkerConfig, vars(args).copy())
    replace_log_level_name(config)
    validate_config_output_mode(config)
    return config


def get_modified_config(parser: ArgumentParser, args: Namespace) -> DarkerConfig:
    """Return configuration options which are set to non-default values"""
    not_default = cast(
        DarkerConfig,
        {
            argument: value
            for argument, value in vars(args).items()
            if value != parser.get_default(argument)
        },
    )
    replace_log_level_name(not_default)
    return not_default


def dump_config(config: DarkerConfig) -> str:
    """Return the configuration in TOML format"""
    dump = toml.dumps(config, encoder=TomlArrayLinesEncoder())
    return f"[tool.darker]\n{dump}"


@dataclass
class Exclusions:
    """File exclusions patterns for pre-processing steps

    For each pre-processor, there is a collection of glob patterns. When running Darker,
    any files matching at least one of the glob patterns is supposed to be skipped when
    running the corresponding pre-processor. If the collection of glob patterns is
    empty, the pre-processor is run for all files.

    The pre-processors whose exclusion lists are currently stored in this data
    structure are
    - Black
    - ``isort``

    """

    black: Collection[str]
    isort: Collection[str]



from itertools import chain
from textwrap import dedent

import pytest

from darker.diff import (
    diff_and_get_opcodes,
    opcodes_to_chunks,
    opcodes_to_edit_linenums,
)
from darker.utils import TextDocument

FUNCTIONS2_PY = dedent(
    """\
    def f(
      a,
      **kwargs,
    ) -> A:
        with cache_dir():
            if something:
                result = (
                    CliRunner().invoke(black.main, [str(src1), str(src2), "--diff", "--check"])
                )
        limited.append(-limited.pop())  # negate top
        return A(
            very_long_argument_name1=very_long_value_for_the_argument,
            very_long_argument_name2=-very.long.value.for_the_argument,
            **kwargs,
        )
    def g():
        "Docstring."
        def inner():
            pass
        print("Inner defs should breathe a little.")
    def h():
        def inner():
            pass
        print("Inner defs should breathe a little.")
"""  # noqa: E501
)


FUNCTIONS2_PY_REFORMATTED = dedent(
    """\
    def f(
        a,
        **kwargs,
    ) -> A:
        with cache_dir():
            if something:
                result = CliRunner().invoke(
                    black.main, [str(src1), str(src2), "--diff", "--check"]
                )
        limited.append(-limited.pop())  # negate top
        return A(
            very_long_argument_name1=very_long_value_for_the_argument,
            very_long_argument_name2=-very.long.value.for_the_argument,
            **kwargs,
        )


    def g():
        "Docstring."

        def inner():
            pass

        print("Inner defs should breathe a little.")


    def h():
        def inner():
            pass

        print("Inner defs should breathe a little.")
    """
)


EXPECT_OPCODES = [
    ("equal", 0, 1, 0, 1),
    ("replace", 1, 3, 1, 3),
    ("equal", 3, 6, 3, 6),
    ("replace", 6, 8, 6, 8),
    ("equal", 8, 15, 8, 15),
    ("insert", 15, 15, 15, 17),
    ("equal", 15, 17, 17, 19),
    ("insert", 17, 17, 19, 20),
    ("equal", 17, 19, 20, 22),
    ("insert", 19, 19, 22, 23),
    ("equal", 19, 20, 23, 24),
    ("insert", 20, 20, 24, 26),
    ("equal", 20, 23, 26, 29),
    ("insert", 23, 23, 29, 30),
    ("equal", 23, 24, 30, 31),
]


def test_diff_and_get_opcodes():
    src = TextDocument.from_str(FUNCTIONS2_PY)
    dst = TextDocument.from_str(FUNCTIONS2_PY_REFORMATTED)
    opcodes = diff_and_get_opcodes(src, dst)
    assert opcodes == EXPECT_OPCODES


def test_opcodes_to_chunks():
    src = TextDocument.from_str(FUNCTIONS2_PY)
    dst = TextDocument.from_str(FUNCTIONS2_PY_REFORMATTED)

    chunks = list(opcodes_to_chunks(EXPECT_OPCODES, src, dst))

    assert chunks == [
        (1, ("def f(",), ("def f(",)),
        (2, ("  a,", "  **kwargs,"), ("    a,", "    **kwargs,")),
        (
            4,
            (") -> A:", "    with cache_dir():", "        if something:"),
            (") -> A:", "    with cache_dir():", "        if something:"),
        ),
        (
            7,
            (
                "            result = (",
                "                CliRunner().invoke(black.main, [str(src1), str(src2), "
                '"--diff", "--check"])',
            ),
            (
                "            result = CliRunner().invoke(",
                '                black.main, [str(src1), str(src2), "--diff", "--check"]',  # noqa: E501
            ),
        ),
        (
            9,
            (
                "            )",
                "    limited.append(-limited.pop())  # negate top",
                "    return A(",
                "        very_long_argument_name1=very_long_value_for_the_argument,",
                "        very_long_argument_name2=-very.long.value.for_the_argument,",
                "        **kwargs,",
                "    )",
            ),
            (
                "            )",
                "    limited.append(-limited.pop())  # negate top",
                "    return A(",
                "        very_long_argument_name1=very_long_value_for_the_argument,",
                "        very_long_argument_name2=-very.long.value.for_the_argument,",
                "        **kwargs,",
                "    )",
            ),
        ),
        (16, (), ("", "")),
        (16, ("def g():", '    "Docstring."'), ("def g():", '    "Docstring."')),
        (18, (), ("",)),
        (
            18,
            ("    def inner():", "        pass"),
            ("    def inner():", "        pass"),
        ),
        (20, (), ("",)),
        (
            20,
            ('    print("Inner defs should breathe a little.")',),
            ('    print("Inner defs should breathe a little.")',),
        ),
        (21, (), ("", "")),
        (
            21,
            ("def h():", "    def inner():", "        pass"),
            ("def h():", "    def inner():", "        pass"),
        ),
        (24, (), ("",)),
        (
            24,
            ('    print("Inner defs should breathe a little.")',),
            ('    print("Inner defs should breathe a little.")',),
        ),
    ]


EXAMPLE_OPCODES = [
    # 0-based, end-exclusive
    ("replace", 0, 4, 0, 1),
    ("equal", 4, 6, 1, 3),
    ("replace", 6, 8, 3, 5),
    ("equal", 8, 15, 5, 12),
    ("insert", 15, 15, 12, 14),
    ("equal", 15, 17, 14, 16),
    ("insert", 17, 17, 16, 17),
    ("equal", 17, 19, 17, 19),
    ("insert", 19, 19, 19, 20),
    ("equal", 19, 20, 20, 21),
    ("insert", 20, 20, 21, 23),
    ("equal", 20, 23, 23, 26),
    ("insert", 23, 23, 26, 27),
    ("equal", 23, 24, 27, 28),
    ("replace", 24, 34, 28, 38),
    ("equal", 34, 35, 38, 39),
]


@pytest.mark.kwparametrize(
    dict(
        context_lines=0,
        multiline_string_ranges=[],
        expect=[0, 3, 4, 12, 13, 16, 19, 21, 22, 26, [28, 37]],  # 0-based
    ),
    dict(
        context_lines=1,
        multiline_string_ranges=[],
        expect=[[0, 5], [11, 23], [25, 38]],  # 0-based, end-inclusive
    ),
    dict(
        context_lines=2,
        multiline_string_ranges=[],
        expect=[[0, 6], [10, 38]],  # 0-based, end-inclusive
    ),
    dict(
        context_lines=0,
        multiline_string_ranges=[  # 0-based, end exclusive
            (2, 4),  # partial left overlap with (3, 5)
            (13, 15),  # partial right overlap with (12, 14)
            (16, 17),  # exact overlap with (16, 17)
            (18, 21),  # overextending overlap with (19, 20)
            (22, 27),  # inner overlap with (21, 23) and full overlap with (26, 27)
            (28, 30),  # full overlap with (28, 38)...
            (36, 46),  # ...partial left overlap with (28, 38)
        ],
        expect=[0, [2, 4], [12, 14], 16, [18, 26], [28, 45]],  # 0-based, end-inclusive
    ),
)
def test_opcodes_to_edit_linenums(context_lines, multiline_string_ranges, expect):
    """`opcodes_to_edit_linenums()` gives correct results"""
    edit_linenums = list(
        opcodes_to_edit_linenums(
            EXAMPLE_OPCODES,
            context_lines,
            # Convert ranges from 0 to 1 based. The test case is defined using 0-based
            # ranges so it's easier to reason about the relation between multi-line
            # string ranges and opcode ranges.
            [(start + 1, end + 1) for start, end in multiline_string_ranges],
        )
    )
    # Normalize expected lines/ranges to 0-based, end-exclusive.
    expect_ranges = [[n, n] if isinstance(n, int) else n for n in expect]
    expect_linenums = list(chain(*(range(n[0], n[1] + 1) for n in expect_ranges)))

    # Normalize result to 0-based, end-exclusive before comparison
    assert [linenum - 1 for linenum in edit_linenums] == expect_linenums


def test_opcodes_to_edit_linenums_empty_opcodes():
    result = list(
        opcodes_to_edit_linenums([], context_lines=0, multiline_string_ranges=[])
    )

    assert result == []
